# cython: language_level=3
"""Simulation of the  Clifford group of up to 12 qubits.

"""

from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

from collections.abc import Iterable
from numbers import Integral, Complex
import cmath, time

import numpy as np
from libc.stdint cimport uint64_t, uint32_t, int32_t, uint8_t, int8_t
from libc.math cimport fabs, log, round
from libc.stdlib cimport malloc, free


cimport cython
include "../pxd_files/clifford12.pxi"
cimport clifford12 as cl
ctypedef qstate12_type *p_qstate12_type
 
 
####################################################################
# Handling errors from C functions in file qstate.c
####################################################################
 

# Messages for error codes returned by the functions in module qstate12.c
QSTATE12_ERROR_STRINGS = {
 -1: "No object with requested property found",
 -2: "Inconststent data in QStateMatrix",
 -3: "Qubit index error at QStateMatrix",
 -4: "State or matrix is too large for a QStateMatrix",
 -5: "Internal buffer overflow in a QStateMatrix",
 -6: "Bit matrix part Q of QStateMatrix is not symmetric",
 -7: "Bad row index for bit matrix in a QStateMatrix",
 -8: "Internal parameter error in a QStateMatrix",
 -9: "Overflow or underflow in scalar factor in a QStateMatrix",
 -10: "A qubit in a ctrl-not gate cannot control itself",
 -11: "Shape mismatch in QStateMatrix comparison",
 -12: "Scalar factor in QStateMatrix is not an integer",
 -13: "General Parameter error",
 -101: "Shape mismatch in QStateMatrix operation",
 -102: "QStateMatrix is not invertible",
 -103: "QStateMatrix is not in the Pauli group",
 -104: "QStateMatrix is not monomial",
 
 # Additional error codes taken from file xsp2co1.c
 -201: "Internal error in operation of group Co_0",
 -202: "Internal Error in operation of group G_x0",
 -203: "Element of G_x0 is not in subroup 2^{1+24}",
 -204: "Bad tag for atom in group G_x0", 
 -205: "Bad element of the group G_x0", 

}



def error_string(code):
    """Map an error code to a string"""
    try:
        return QSTATE12_ERROR_STRINGS[code]
    except KeyError:
        err = "Internal error %d in processing QState12 instance"
        return err % code

cpdef int32_t chk_qstate12(int32_t code) except -1:
    """Check the return code of a function in module qstate12.c
    
    The function raises ValueError with a suitable error message if
    the return value ``code``  of a function in module ``qstate12.c``
    is negative. If ``code`` is nonnegative then the function
    returns ``code``.
    """
    if code >= 0:
        return code
    raise ValueError(error_string(code))


####################################################################
# Class QStateMatrix 
####################################################################


cdef class QState12(object):
    """TODO: Yet to be documented!!!"""

    cdef uint64_t data_[QSTATE12_MAXROWS] # data storage
    cdef qstate12_type qs     # structure to store a state vector 
    
    #########################################################################
    # Constructors

    def __cinit__(self, *args, **kwds):
        """Initialize a quadratic state tensor to the scalar 0."""
        cl.qstate12_set_mem(&self.qs, &(self.data_[0]), QSTATE12_MAXROWS);
        cl.qstate12_zero(&self.qs, 0)

    def __init__(self, rows, cols = None, data = None, mode = 0):
        """Intitalize a quadratic state matrix
        
        If ``rows`` is an integer then ``cols`` must also be
        an integer and a quadratic state matrix of shape        
        ``(rows, cols)`` is created. The parameter ``data`` may be:

            * ``None`` (default). Then the zero matrix is created.
            
            * A list of integers. Then that list of integers must 
              encode a valid pair ``(A, Q)`` of bit matrices that make
              up the quadratic state matrix. In this case parameter 
              ``mode`` is  evaluated as follows:
              
               * 1: create matrix ``Q`` from lower triangular part
              
               * 2: create matrix ``Q`` from upper triangular part
               
               * Anything else: matrix ``Q`` must be symmetric.
               
              In this case ``self.factor`` is set to  one.
              
            * An integer ``v``. Then one of the values ``rows``
              and ``cols`` must be zero and the unit (row or
              column) vector with index ``v`` is created.
              This corresponds to a complex matrix with one 
              row or one column.  

        If ``rows`` is an instance of class QState12 then a deep
        copy of that instance is create.                
        """
        cdef uint32_t n, n0, n1, nrows, i
        cdef uint64_t v
        if isinstance(rows, Integral) and isinstance(rows, Integral):
            n0, n1 = rows, cols
            n = n0 + n1
            if data is None:
                chk_qstate12(cl.qstate12_zero(&self.qs, n))                
            elif isinstance(data, Iterable):
                nrows = len(data)
                self.qs.reduced = 0
                for i in range(min(nrows, QSTATE12_MAXROWS)):
                    self.data_[i] = data[i]
                chk_qstate12(cl.qstate12_set(
                    &self.qs, n, nrows, &self.data_[0], mode))
            elif isinstance(data, Integral) and rows * cols == 0:
                v = data
                chk_qstate12(cl.qstate12_vector_state(
                    &self.qs, n, v))
            else:
                err = "Bad data type for QStateMatrix"
                raise TypeError(err) 
            chk_qstate12(cl.qstate12_mat_reshape(&self.qs, n0, n1))
        elif isinstance(rows, QState12) and cols is None:
            source_pqs = pqs12(rows)
            chk_qstate12(cl.qstate12_copy(source_pqs, &self.qs))
        else:
            err = "Cannot construct QStateMatrix from given input" 
            raise TypeError(err) 



    #########################################################################
    # Properties and methods for obtaining implementation details

    property data:
        """Get a copy of the data bit matrix of the state as a list"""
        def __get__(self):
            """Return a copy of the data bits of the  state as a numpy array"""
            chk_qstate12(cl.qstate12_reduce(&self.qs))
            chk_qstate12(cl.qstate12_check(&self.qs))
            cdef uint32_t i
            cdef uint32_t imax = self.qs.nrows
            cdef uint64_t *a = &self.data_[0]
            return  [int(a[i]) for i in range(imax)]
            
    property raw_data:
        """Get a copy of the raw data bit matrix of the state as a list
        
        For debugging and testing only!
        """
        def __get__(self):
            cdef uint32_t i
            cdef uint32_t imax = QSTATE12_MAXROWS
            cdef uint64_t *a = &self.data_[0]
            return  [int(a[i]) for i in range(imax)]
        
    property nrows:
        """Get the number of rows of the bit matrix of the state """
        def __get__(self):
            return self.qs.nrows 

    property ncols:
        """Get the number of columns the bit matrix A of the state 
        
        ``self.ncols == n`` means that the state ``qs`` referred by
        ``self`` is a complex ``2**n`` dimensional vector describing
        a stabilizer state of ``n`` qubits.        
        """
        def __get__(self):
            return self.qs.ncols 

    property factor:
        """Get the factor of the state 
        
        The factor is returned as a pair ``(e, phi)`` which denotes
        the factor :math:`2^{e/2} \cdot \exp(\phi \pi i /4)` .           
        """
        def __get__(self):
            cdef int64_t r = self.qs.nrows
            cdef int64_t f = self.qs.factor
            cdef int64_t e = (f & -0x10) >> 4 if r else 0
            cdef int64_t phi = f & 7 if r else 0
            return (e, phi)


                                
    def row(self, uint32_t i):    
        """Return row i of the bit matrix of the state"""
        self.check()
        cdef uint64_t v
        if i >= self.qs.nrows:
            chk_qstate12(-7)
        v = self.data_[i]
        return int(v)
        
    def column(self, uint32_t j):  
        """Return column j of the bit matrix of the state"""
        self.check()
        return int(cl.qstate12_get_column(&self.qs, j))
            
    def mul_av(self, uint64_t v):
        """Return matrix product ``A * transposed(v)``
        
        Here ``v`` is an integer denoting a bit vector and ``A`` is the 
        ``A`` of the bit matrix of the state.
        
        The result is returned as an integer denoting a bit matrix.
        """
        cdef uint64_t w
        chk_qstate12(cl.qstate12_mul_Av(&self.qs, v, &w))
        return int(w)


    #########################################################################
    # Elementary operations and checks
    
    def is_reduced(self):
        return bool(self.qs.reduced)

    property shape:
        """Get shape of the complex matrix represented by the state 
        
        The function returns a pair ``(rows, cols)`` meaning
        that the state corresponds to a complex
        ``2**nrows`` times ``2**ncols`` matrix.        
        """
        def __get__(self):
            cdef uint32_t cols = self.qs.shape1
            cdef uint32_t rows = self.qs.ncols - cols
            return rows, cols  

    def copy(self):
        return self.__class__(self)

    def conjugate(self):
        """Conjugate state in place"""
        chk_qstate12(cl.qstate12_conjugate(&self.qs))
        return self
        
    def mul_scalar(self, int32_t e, int32_t phi=0):
        """Multiply the state by a sclar factor in place
        
        The factor is :math:`2^{e/2} \cdot \exp(\phi \pi i /4)` .           
        """
        chk_qstate12(cl.qstate12_mul_scalar(&self.qs, e, phi & 7))
        return self
        
    def transpose(self):
        """Transpose state matrix im place"""
        chk_qstate12(cl.qstate12_mat_t(&self.qs))
        return self

    def lb_rank(self):
        """Return binary logarithm of the rank of a matrix.

        Return -1 if that rank is zero.
        """
        cdef int32_t res = cl.qstate12_mat_lb_rank(&self.qs)
        if (res >= -1):
            return res
        chk_qstate12(res)

    def trace(self):
        """Return the trace of a square matrix.

        The trace is returned as an integer, a floating point
        number, or a complex number.
        """
        cdef double tr[2]
        cdef double[:] tr_view = tr
        cdef int32_t res = cl.qstate12_mat_trace(
            &self.qs, &tr_view[0])
        if res == 4:
            return tr[0] + 1j * tr[1]
        if (res >= 2 or abs(tr[0]) > (1 << 64)):
            return tr[0]
        return int(tr[0])
        

    def _trace_factor(self):
        cdef int32_t factor = 0
        cdef int32_t res = cl.qstate12_mat_trace_factor(
            &self.qs, &factor)
        return res, factor

    def set_zero(self):
        """Set state matrix to zero"""
        self.qs.factor = self.qs.nrows = self.reduced = 0
        return self

    def check(self):
        """Raise ValueError is state matrix is bad""" 
        cl.qstate12_check(&self.qs)
        return self

    def check_code(self):
        """Check a state
        
        Return ``0`` if state is ok.
        
        If the state is bad then return an error code as a 
        negative integer.
        """ 
        cdef int32_t res = cl.qstate12_check(&self.qs)
        return res 

    def monomial_row_op(self):
        a = np.zeros(self.qs.nrows, dtype = np.uint32)
        cdef uint32_t[:] a_view = a 
        cdef int32_t res
        res = cl.qstate12_monomial_matrix_row_op(&self.qs, &a_view[0])
        if (res >= 0):
            return a
        chk_qstate12(res)

    #########################################################################
    # Reshaping a state matrix
    
    def reshape(self, int32_t rows, int32_t cols):
        chk_qstate12(cl.qstate12_mat_reshape(&self.qs, rows, cols))
        return self

    #########################################################################
    # Reducing a state matrix to a standard form 

    def echelon(self):
        """Convert state to (not reduced) echelon form

        Only the first  ``n_reduce`` colums are converted to reduced
        echelon form

        The function reutrns a list ``table`` with ``table[j] = i``
        if the leading coefficient of row ``i`` is in columng ``j``.
        We have  ``table[j] = -1``  if no such row exists.
        """
        chk_qstate12(cl.qstate12_echelonize(&self.qs))
        return self

    def reduce(self, forced = False):
        """Reduce a state to a standard form
        
        The representation of a state is not unique.  This function
        changes a sate to a unique standard form where the bit matrix A
        of the internal representation of the state is in reduced
        echelon form, as described in the guide.
        
        If ``forced`` is ``True``, also a previously reduced state
        is recuced one again.
        """
        if forced:
            self.qs.reduced = 0
        chk_qstate12(cl.qstate12_reduce(&self.qs))
        return self 

    def reduce_matrix(self):
        """Yet to be documented"""
        cdef uint8_t row_table[QSTATE12_MAXCOLS+1]
        chk_qstate12(cl.qstate12_reduce_matrix(&self.qs, &row_table[0]))
        cdef uint32_t i = self.nrows + self.ncols
        cdef uint32_t j
        return [ row_table[j] for j in range(i) ]

 
                

    #########################################################################
    # Permuting the qubits of the state

    def rot_bits(self, int32_t rot, uint32_t nrot, uint32_t n0 = 0):  
        """Wrapper for the corresponding C function"""
        chk_qstate12(cl.qstate12_rot_bits(&self.qs, rot, nrot, n0))
        return self
        

    def xch_bits(self, uint32_t sh, uint64_t mask):
        """Wrapper for the corresponding C function"""
        chk_qstate12(cl.qstate12_xch_bits(&self.qs, sh, mask))
        return self

    #########################################################################
    # Applying qubit gates

    def gate_not(self, uint64_t v):
        """Wrapper for the corresponding C function"""
        chk_qstate12(cl.qstate12_gate_not(&self.qs, v))
        return self

    def gate_ctrl_not(self, uint64_t vc, uint64_t v):
        """Wrapper for the corresponding C function"""
        chk_qstate12(cl.qstate12_gate_ctrl_not(&self.qs, vc, v))
        return self

    def gate_phi(self, uint64_t v, uint32_t phi):
        """Wrapper for the corresponding C function"""
        chk_qstate12(cl.qstate12_gate_phi(&self.qs, v, phi))
        return self

    def gate_ctrl_phi(self, uint64_t v1, uint64_t v2):
        """Wrapper for the corresponding C function"""
        chk_qstate12(cl.qstate12_gate_ctrl_phi(&self.qs, v1, v2))
        return self

    def gate_h(self,  uint64_t v):
        """Wrapper for the corresponding C function"""
        chk_qstate12(cl.qstate12_gate_h(&self.qs, v))
        return self
        
    #########################################################################
    # Extending and restricting a state

    def extend_zero(self, uint32_t j, uint32_t nqb):
        """Wrapper for the corresponding C function"""
        chk_qstate12(cl.qstate12_extend_zero(&self.qs, j, nqb))
        return self

    def extend(self, uint32_t j, uint32_t nqb):
        """Wrapper for the corresponding C function"""
        chk_qstate12(cl.qstate12_extend(&self.qs, j, nqb))
        return self
        
    def restrict_zero(self, uint32_t j, uint32_t nqb):
        """Wrapper for the corresponding C function"""
        chk_qstate12(cl.qstate12_restrict_zero(&self.qs, j, nqb))
        return self

    def restrict(self, uint32_t j, uint32_t nqb):
        """Wrapper for the corresponding C function"""
        chk_qstate12(cl.qstate12_restrict(&self.qs, j, nqb))
        return self

    def sumup(self, uint32_t j, uint32_t nqb):
        """Wrapper for the corresponding C function"""
        chk_qstate12(cl.qstate12_sum_cols(&self.qs, j, nqb))
        return self
        
    #########################################################################
    # Conversion of a state matrix to a complex vector
        
    def complex(self):
        """Convert the state to a complex vector"""
        cdef uint32_t n0, n1
        n0, n1 = self.shape
        a = np.empty(2 << self.ncols, dtype = np.double)
        cdef double[:] a_view = a
        chk_qstate12(cl.qstate12_complex(&self.qs, &a_view[0]))
        c = a[0::2] + 1j * a[1::2]
        del a
        return c.reshape((1 << n0, 1 << n1))

    #########################################################################
    # Comparing state matrices

    def __eq__(self, QState12 other):
        """Return True iff two states are equal"""
        cdef p_qstate12_type other_pqs
        if isinstance(other, QState12):
            other_pqs = pqs12(other)
            return chk_qstate12(cl.qstate12_equal(&self.qs, other_pqs))
        err = "Cannot compare QStateMatrix with non QStateMatrix"
        raise TypeError(err)
        
        
    
    #########################################################################
    # Matrix multiplication and inversion
       
    def __imatmul__(self, QState12 other):
        """Perform a matrix multiplication.
        
        We replace ``self`` by the matrix product ``self @ other``.
        ``other`` is not changed.        
        """
        cdef int32_t res
        cdef p_qstate12_type pqs2
        if isinstance(other, QState12):
            pqs2 = pqs12(other)
            res = cl.qstate12_matmul(&self.qs, pqs2, &self.qs)
            if res >= 0:
                return self
            err = "Multiplying QStateMatrix objects of shape %s and %s"
            print("\n" + err % (self.shape, other.shape))
            chk_qstate12(res)
        err = "Bad type for matrix multiplication with QStateMatrix"
        raise TypeError(err)

    def __matmul__(self, QState12 other):
        """Return matrix product  ``self @ other``.
        
        """
        return self.copy().__imatmul__(other)

    def _mat_inv(self):
        """Return inverse matrix"""
        qs1 = self.copy()
        cdef p_qstate12_type pqs1 = pqs12(qs1)
        chk_qstate12(cl.qstate12_mat_inv(pqs1))
        return qs1

    #########################################################################
    # Scalar multiplication 


    def  __imul__(self, other):
        cdef p_qstate12_type pqs
        cdef uint32_t nqb, shape1
        if isinstance(other, Complex):
            qstate12_mul_scalar(self, other)
            return self
        elif isinstance(other, QState12):
            if self.shape == other.shape:
                pqs = pqs12(other)
                nqb = self.ncols
                shape1 = self.qs.shape1
                chk_qstate12(cl.qstate12_product(&self.qs, pqs, nqb, 0))
                self.qs.shape1 = shape1
                return self
            else:
                err = "QStateMatrix instances must have same shape"
                raise ValueError(err)
        else:
            err = "Bad type for multiplication with QStateMatrix"
            raise TypeError(err)


    def  __mul__(self, other):
        return self.copy().__imul__(other)
        
    __rmul__ = __mul__    

    def __itruediv__(self, other):
        if isinstance(other, Complex):
            return self.__imul__(1.0/other)
        else:
            err = "Bad type for multiplication with QStateMatrix"
            raise TypeError(err)

    def __truediv__(self, other):
        return self.copy().__itruediv__(other)
            
    def __neg__(self):
        return self.copy().mul_scalar(0, 4)    

    def __pos__(self):
        return self    

    
    #########################################################################
    # A rather general multiplication function

    def qstate12_product(self, QState12 other, uint32_t nqb, uint32_t nc):
        """Wrapper for the corresponding C function
    
        We return a certain product ``qs3`` of the states ``qs1`` 
        ad ``qs2`` referred by ``self`` and ``other``.
        
        Let ``n1 = qs1.ncols,  n2 = qs2.ncols``.
        Put ``qs1a =  qs1.extend(n1, n2-nqb)``,
        ``qs2a =  qs2.extend(nqb, n1-nqb)``. Then ``qs1a`` and
        ``qs2a`` are complex functions on ``(nn1 + nn2 - nqb)``
        qubits. Let ``qs3a`` be the complex function which is the 
        product of the functions  ``qs1a`` and ``qs2a``. Then we 
        have ``qs3 = qs3a.sum_cols(0, nc)``. So we have
        ``qs3.ncols = nn1 + nn2 - nqb - nc``.
        The result ``qs3`` is computed in the state referred by 
        ``qs1``. It is reduced. 
 
        E.g. ``qs1.product(qs2, nc, nc)`` is the tensor 
        contraction over the first ``nc`` qubits of ``qs1`` and 
        ``qs2``. ``In case qs1.ncols = qs2.ncols = n``, the 
        function ``qs1.product(qs2, 0, n)`` returns the 
        product of ``qs1`` and ``qs2`` (considered as functions); 
        and ``qs1.product(qs2, n, n)`` returns the scalar
        product of ``qs1`` and ``qs2`` (considered as vectors).
 
        ``qs1.product(qs2, 0, n)`` corresponds to the 
        function ``(qs1 (*) qs2)_n``  where ``(*)`` denotes the 
        ``'\odot'`` operation defined in section 
        *Products and tensor products of quadratic mappings*
        of the guide.   
        """
        qs1 = self.copy()
        cdef p_qstate12_type pqs1 = pqs12(qs1)
        cdef p_qstate12_type pqs2 = pqs12(other)
        chk_qstate12(cl.qstate12_product(pqs1, pqs2, nqb, nc))
        return qs1

    #########################################################################
    # Auxiliary function for Matrix multiplication (for testing)

    def qstate12_prep_mul(self, QState12 other, uint32_t nqb):
        """Wrapper for the corresponding C function
    
        To be used for tests only. Function ``qstate12_prep_mul``
        is an auxliary function for function ``qstate12_product``.    
        """
        qs1 = self.copy()
        qs2 = other.copy()
        cdef p_qstate12_type pqs1 = pqs12(qs1)
        cdef p_qstate12_type pqs2 = pqs12(qs2)
        row_pos = chk_qstate12(cl.qstate12_prep_mul(pqs1, pqs2, nqb))
        return row_pos, qs1, qs2
        

    #########################################################################
    # Obtain complex entries of a state
    
    @cython.boundscheck(False)    
    def entries(self, indices):
        """Return complex entries of a state.
        
        Here ``indices`` must be a one-dimensional numpy array of 
        indices of ``dtype = np.uint32``. Unused bits of the 
        indices are ignored.
        
        The function returns a complex numpy array ``c`` of the same 
        shape as the input so that ``c[i]`` is the complex value
        of entry ``indices[i]`` of the state.
              
        This function does not change the state. It uses method 
        ``echelonize`` to bring the representation of the
        state to echelon form.
        """
        cdef uint32_t[:] ind_view = indices
        cdef unsigned int n = len(indices)
        a = np.empty(2 * n, dtype = np.double, order = 'C')
        cdef double[:] a_view = a
        cdef int32_t res = 0
        if n:
            res = chk_qstate12(cl.qstate12_entries(
                &self.qs, n, &ind_view[0], &a_view[0]))
        if res == 4:
            c = a[0::2] + 1j * a[1::2] 
            del a
        else:
            c = a[0::2] 
        return c

    #########################################################################
    # Some matrix functions
    
    def pauli_vector(self):
        """TODO: yet to be documented!!!"""
        cdef uint64_t v;
        chk_qstate12(cl.qstate12_pauli_vector(&self.qs, &v))
        return int(v)
        
    def pauli_conjugate(self, v, arg = True):   
        v = np.array(v, dtype = np.uint64, copy=True)
        shape = v.shape
        assert len(shape) <= 1
        v = np.ravel(v)
        if len(v) == 0:
            return []
        cdef uint64_t[:] v_view = v
        chk_qstate12(cl.qstate12_pauli_conjugate(
            &self.qs, len(v), &v_view[0], arg))
        if len(shape):
            return [int(x) for x in v]
        else:
            return int(v[0])

    def to_symplectic(self):
        """Convert a quadratic state matrix to a symplectic bit matrix

        Documentation see corresponding method in subclass
        ``mmgroup.structures.qs_matrix.QStateMatrix``.
        """
        a = np.zeros(32, dtype = np.uint64)
        cdef uint64_t[:] a_view = a
        cdef int32_t res = chk_qstate12(
            cl.qstate12_to_symplectic(&self.qs, &a_view[0]))
        return a[:res]

 
####################################################################
# Auxiliary functions  
####################################################################

 
cdef p_qstate12_type pqs12(QState12 state):
    """Return pointer to the structure of a ``state`` 
    
    Here a ``state`` must be of type ``QState12``.
    """
    return &state.qs
  
 
####################################################################
# Convert a complex number to a factor for class QStateMatrix 
####################################################################
    

MIN_ABS = 2.0**-1024
EPS = 1.0e-8 
PI =  3.141592653589793238462643383279
LOG2 = 0.693147180559945309417232121458

       
def qstate12_mul_scalar(QState12 qs, x):
    """Multiply quadratic state matrix with scalar factor
    
    Here ``qs`` is an instance of class ``QState12``
    representing a quadratic state matrix and ``x`` is
    a complex number.
    
    This multiplication succeeds if ``x = 0`` or 
    ``x == 2**(0.5*e) * z**phi``, with ``z = (1+1j)/sqrt(2)``
    and integers ``e, phi``.
    
    Raise ValueError multiplication fails.. We accept a relative 
    error of about ``1.0e-8`` for ``x``.
    
    So it is safe to write e.g. ``2**-0.5 * (-1+1j) * m`` 
    if ``m`` is an  instance of class ``QState12``.    
    """
    cdef double r, phi, e, e_r, phi8_r
    cdef int32_t e_i, phi8_i
    r, phi = cmath.polar(x)
    if r <= MIN_ABS:
        return qs.set_zero()
    e = 2.0 * log(r) / LOG2
    phi8 = 4.0 * phi / PI
    e_r, phi8_r = round(e), round(phi8)
    e_i, phi8_i = <int32_t>e_r, <int32_t>phi8_r
    if (max(fabs(e - e_r), fabs(phi8 - phi8_r)) < EPS):
        return qs.mul_scalar(e_i, phi8_i)
    err = "Cannot convert number to factor for QStateMatrix"
    raise ValueError(err)    




 
####################################################################
# Wrappers for C functions contructing state matrices
####################################################################

def qstate12_unit_matrix(QState12 qs, uint32_t n):
    """Change state qs to a  2**n times 2**n unit matrix"""
    cdef p_qstate12_type m_pqs = pqs12(qs)
    chk_qstate12(cl.qstate12_unit_matrix(m_pqs, n))
    return qs


def qstate12_pauli_matrix(QState12 qs, uint32_t n, uint64_t v):
    """Change state qs to a  2**n times 2**n Pauli matrix"""
    cdef p_qstate12_type m_pqs = pqs12(qs)
    chk_qstate12(cl.qstate12_pauli_matrix(m_pqs, n, v))
    return qs


def qstate12_column_monomial_matrix(QState12 qs, uint32_t nqb, a):
    """Change state ``qs`` to a monomial matrix

    Set the state ``qs`` to a real monomial 
    ``2**nqb`` times ``2**nqb`` transformation matrix ``T`` which 
    is monomial in  the sense that each column contains exactly 
    one nonzero entry ``1`` or ``-1``. So left multiplication  
    with ``T`` maps maps unit vectors to (possibly negated) unit  
    vectors. It transforms a  state ``<v|`` of ``nqb`` input bits 
    to a state ``T * <v|`` of ``nqb`` output qubits. Here
    ``|v>`` is considered as a column vector.

    ``pa`` refers to an array a of integers ``a[i]`` of length 
    ``nqb + 1``. Each integer ``a[i]`` is interpreted as a bit 
    field  via its binary  representation. So ``a[i,j]`` means 
    ``(a[i] >> j) & 1``. ``a[i, j1:j2]`` means the bit field  
    ``a[i,j1],...,a[i,j2-1]``.
   
    For any bit vector ``v`` of length ``nqb`` let ``|v>`` be the 
    unit vector with index ``v``. For any bit vector ``v`` of 
    length ``nqb + 1`` let ``|v>`` be the (possibly negated) unit 
    vector ``(-1)**v[nqb] * |v[0:nqb]>``.  ``|v1 ^ v2>`` and 
    ``|1 << v1>`` are defined via the corrresponding operators 
    ``<<`` and ``^`` in C.
  
    Then ``T``  maps
  
      * ``|0>``     to  ``|a[0, 0:nqb+1]>``
  
      * ``|1 << i>`` to  ``|a[0, 0:nqb+1] ^ a[i+1, 0:nqb+1]>``
  
    ``T`` maps unit vectors to (possibly negated) unit vectors, 
    so ``T(v)`` is well defined by ``|T(v)> = T(|v>)`` for a bit 
    field ``v`` of length ``nqb + 1``. We have
 
      * ``|T(v1 ^ v2)> = (-1)**Q(v1,v2) * |T(v1) ^ T(v2) ^ T(0)>``,
 
    for bit fields ``v1, v2`` of length ``nqb + 1`` and an 
    alternating bilinear form ``Q`` depending on the first ``nqb`` 
    bits of ``v1`` and ``v2`` only. Thus ``T`` is  defined by the 
    above equation and ``Q``. The bilinear form ``Q`` is defined by:
     
      * ``Q(v1, v2) = Q(v2, v1),  Q(v1, v1) = 0``,  and
 
      * ``Q(1 << i, 1 << j) =  m[i + 1, j + nqb + 1]``,  
        for ``0 <= j < i < nqb``.
    """
    cdef p_qstate12_type m_pqs = pqs12(qs)
    cdef uint64_t aa[QSTATE12_MAXROWS+1]
    if nqb >= QSTATE12_MAXROWS:
        return chk_qstate12(-4) 
    cdef int i
    for i in range(nqb + 1):
        aa[i] = a[i]
    chk_qstate12(cl.qstate12_monomial_column_matrix(m_pqs, nqb, &aa[0])) 
    return qs


def qstate12_row_monomial_matrix(QState12 qs, uint32_t nqb, a):
    """Change state ``qs`` to a monomial matrix

    Equivalent to 

    ``qstate12_column_monomial_matrix(qs, nqb, a)``
    ``qs = qs.T``
    """
    cdef p_qstate12_type m_pqs = pqs12(qs)
    cdef uint64_t aa[QSTATE12_MAXROWS+1]
    if nqb >= QSTATE12_MAXROWS:
        return chk_qstate12(-4) 
    cdef int i
    for i in range(nqb + 1):
        aa[i] = a[i]
    chk_qstate12(cl.qstate12_monomial_row_matrix(m_pqs, nqb, &aa[0])) 
    return qs

####################################################################
# Wrappers for other exported C functions  
####################################################################



def qstate12_pauli_vector_mul(uint32_t nqb, uint64_t v1, uint64_t v2):
    if nqb > 31:
        raise ValueError("Bad Pauli group vector")
    return int(cl.qstate12_pauli_vector_mul(nqb, v1, v2))

def qstate12_pauli_vector_exp(uint32_t nqb, uint64_t v, uint32_t e):
    if nqb > 31:
        raise ValueError("Bad Pauli group vector")
    return int(cl.qstate12_pauli_vector_exp(nqb, v, e))


####################################################################
# Support for the subgroup 2^{1+24}.Co_1 of the monster  
####################################################################


@cython.wraparound(False)
@cython.boundscheck(False)
def xsp2co1_chain_short_3(QState12 qstate, src, dest):
    cdef uint64_t[:] src_view = src
    cdef uint64_t[:] dest_view = dest
    cdef p_qstate12_type pqs = pqs12(qstate)
    cdef uint32_t length = len(src)
    assert len(dest) >= length
    if length > 0:
        chk_qstate12(cl.xsp2co1_chain_short_3(pqs, length, &src_view[0],
            &dest_view[0]))

@cython.wraparound(False)
@cython.boundscheck(False)
def xsp2co1_elem_to_qs_i(elem):
    assert len(elem) >= 26 
    cdef uint64_t e[26]
    cdef uint32_t i
    for i in range(26): e[i] = elem[i]
    result = QState12(12, 12)
    cdef qstate12_type qs0
    cdef p_qstate12_type pqs = pqs12(result)
    chk_qstate12(cl.xsp2co1_elem_to_qs_i(e, &qs0))
    chk_qstate12(cl.qstate12_copy(&qs0, pqs))
    return result

@cython.wraparound(False)
@cython.boundscheck(False)
def xsp2co1_elem_to_qs(elem):
    cdef uint64_t[:] elem_view = elem
    result = QState12(12, 12)
    cdef p_qstate12_type pqs = pqs12(result)
    chk_qstate12(cl.xsp2co1_elem_to_qs(&elem_view[0], pqs))
    return result


@cython.wraparound(False)
@cython.boundscheck(False)
def xsp2co1_qs_to_elem_i(QState12 qstate, uint64_t x1):
    cdef p_qstate12_type pqs = pqs12(qstate)
    result = np.zeros(26, dtype = np.uint64)
    cdef uint64_t[:] result_view = result
    chk_qstate12(cl.xsp2co1_qs_to_elem_i(pqs, x1, &result_view[0]))
    return result


####################################################################
# Wrappers for other exported bit matrix functions  
####################################################################


@cython.wraparound(False)
@cython.boundscheck(False)
def bitmatrix64_add_diag(m, uint32_t j):
    cdef uint64_t[:] m_view = m
    cdef uint32_t i = len(m)
    cl.bitmatrix64_add_diag(&m_view[0], i, j)


@cython.wraparound(False)
@cython.boundscheck(False)
def bitmatrix64_mask_rows(m, uint64_t mask):
    cdef uint64_t[:] m_view = m
    cdef uint32_t i = len(m)
    cl.bitmatrix64_mask_rows(&m_view[0], i, mask)


@cython.wraparound(False)
@cython.boundscheck(False)
def bitmatrix64_find_masked_row(m, uint64_t mask, uint64_t v):
    cdef uint64_t[:] m_view = m
    cdef uint32_t i = len(m)
    cdef int32_t res
    res = cl.bitmatrix64_find_masked_row(&m_view[0], i, mask, v)
    return res


@cython.wraparound(False)
@cython.boundscheck(False)
def bitmatrix64_t(m, uint32_t ncols):
    m1 = np.array(m, dtype = np.uint64)
    m2 = np.zeros(ncols, dtype = np.uint64)
    cdef uint64_t[:] m1_view = m1
    cdef uint64_t[:] m2_view = m2
    if len(m1) and ncols:
        chk_qstate12(cl.bitmatrix64_t(
            &m1_view[0], len(m1), ncols, &m2_view[0]))
    return m2        


@cython.wraparound(False)
@cython.boundscheck(False)
def bitmatrix64_cap_h(m1, m2, uint32_t j0, uint32_t n):
    cdef uint64_t[:] m1_view = m1
    cdef uint64_t[:] m2_view = m2
    cdef uint32_t i1 = len(m1)
    cdef uint32_t i2 = len(m2)
    cdef uint32_t rows1 = cl.bitmatrix64_echelon_h(&m1_view[0], i1, j0, n)
    cdef uint32_t rows2 = cl.bitmatrix64_echelon_h(&m2_view[0], i2, j0, n)
    cdef uint32_t cap = cl.bitmatrix64_cap_h(&m1_view[0], &m2_view[0], 
         i1, i2, j0, n)
    chk_qstate12(cap)
    return rows1 - cap, rows2 - cap


@cython.wraparound(False)
@cython.boundscheck(False)
def bitmatrix64_mul(m1, m2):
    cdef uint64_t[:] m1_view = m1
    cdef uint64_t[:] m2_view = m2
    cdef uint32_t i1 = len(m1)
    cdef uint32_t i2 = len(m2)
    m3 = np.zeros(i1, dtype = np.uint64)
    cdef uint64_t[:] m3_view = m3
    cl.bitmatrix64_mul(&m1_view[0], &m2_view[0], i1, i2,
        &m3_view[0])
    return m3

@cython.wraparound(False)
@cython.boundscheck(False)
def bitmatrix64_rot_bits(m, int32_t rot, uint32_t nrot, uint32_t n0):
    cdef uint64_t[:] m_view = m
    cdef uint32_t i = len(m)
    cdef uint32_t res = cl.bitmatrix64_rot_bits(&m_view[0], i, rot, nrot, n0)
    chk_qstate12(res)


@cython.wraparound(False)
@cython.boundscheck(False)
def bitmatrix64_xch_bits(m, uint32_t sh, uint64_t mask):
    cdef uint64_t[:] m_view = m
    cdef uint32_t i = len(m)
    cdef uint32_t res = cl.bitmatrix64_xch_bits(&m_view[0], i, sh, mask)
    chk_qstate12(res)

@cython.wraparound(False)
@cython.boundscheck(False)
def bitmatrix64_reverse_bits(m, uint32_t n, uint32_t n0):
    cdef uint64_t[:] m_view = m
    cdef uint32_t i = len(m)
    cdef uint32_t res = cl.bitmatrix64_reverse_bits(&m_view[0], i, n, n0)
    chk_qstate12(res)


    
@cython.wraparound(False)
@cython.boundscheck(False)
def bitmatrix64_inv(m):
    cdef uint32_t i = len(m), j
    m1 = np.zeros(i, dtype = np.uint64)
    for j in range(i):
        m1[j] = m[j]   
    cdef uint64_t[:] m1_view = m1
    cdef uint32_t res = cl.bitmatrix64_inv(&m1_view[0], i)
    chk_qstate12(res)
    return m1


@cython.wraparound(False)
@cython.boundscheck(False)
def bitmatrix32_test_sort(uint32_t alg, a, uint32_t n_rept):
    cdef uint32_t n = len(a), i, j
    cdef uint32_t[:] a_view = a
    cdef uint32_t *b = <uint32_t *> malloc(n * sizeof(uint32_t));
    if not b:
        raise MemoryError()
    cdef double t0, t1
    t0 = time.perf_counter()
    for i in range(n_rept):
        for j in range(n): b[j] = a_view[j]
        if alg == 1: clifford12.bitvector32_sort(b, n)
        elif alg == 2: clifford12.bitvector32_heapsort(b, n)
    t1 = time.perf_counter()
    return t1 - t0
            
    
    






from mmgroup.mm_space import MMSpace, MMV, MMVector
from mmgroup.mm_space import characteristics
from mmgroup.structures.mm0_group import MM0Group

from mmgroup.tests.spaces.sparse_mm_space import SparseMmSpace
from mmgroup.tests.spaces.sparse_mm_space import SparseMmV
from mmgroup.tests.groups.mgroup_n import MGroupN


#print("module mmgroup.tests.spaces.spaces is deprecated!!")


spaces_dict = {}

g = MM0Group()
ref_g = MGroupN()


class TestSpace:
    def __init__(self, p):
        self.p = p
        self.space = MMV(p) 
        self.ref_space = SparseMmV(p)
        self.group = g     
        self.ref_group = ref_g 
    def __call__(self, *args):
        return self.space(*args)



def MMTestSpace(p):
    global spaces_dict
    try:
        return spaces_dict[p]
    except KeyError:
        spaces_dict[p] = TestSpace(p)
        return spaces_dict[p]





from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals



from random import randint #, shuffle, sample
from functools import reduce
from operator import __or__

import numpy as np
import pytest

from mmgroup.structures.qs_matrix import QStateMatrix, qs_rand_matrix
from mmgroup.structures.qs_matrix import qs_unit_matrix
from mmgroup.structures.qs_matrix import flat_product

from mmgroup.tests.test_clifford.test_qs_matrix import compare_complex


#####################################################################
# 
#####################################################################


def prep_mul(m1, m2, nqb):
    row_pos, m1, m2 = m1.qstate12_prep_mul(m2, nqb)
    return row_pos, m1, m2            


qs_matrix_data = [
    [ (0,0, (1,0), []),  (0,0, (1,0), []), 0 ],
    [ (0,4, (0,0), [0b00_0101, 0b00_0011]),
        (0,4, (0,0), [0b00_0111, 0b_00_1001]), 2],
]


def create_m(rows, cols, factor, data):
    """Create ``QStateMatrix`` object from input data"""
    m = QStateMatrix(rows, cols, data)
    m.mul_scalar(*factor) 
    return m



def rand_mul_scalar(m):
    m.mul_scalar(randint(-16, 15), randint(-8,7))

def create_product_testvectors():
    """yield instances of class ``QStateMatrix`` for tests """
    for m1_data, m2_data, nqb in qs_matrix_data:
        m1 = create_m(*m1_data)
        m2 = create_m(*m2_data)
        yield m1, m2, nqb
    yield qs_unit_matrix(4), qs_unit_matrix(4), 3 
    for i in range(20):
        nqb = randint(0,2)
        yield qs_rand_matrix(2, 0, 3), qs_rand_matrix(2, 0, 3), nqb
    for cols1 in range(2,6):
        for cols2 in range(2,6):
            for nqb in range(min(cols1, cols2)):
               for n in range(2):
                   m1 = qs_rand_matrix(0, cols1, cols1+3)
                   rand_mul_scalar(m1)  
                   m2 = qs_rand_matrix(0, cols2, cols2+3)
                   rand_mul_scalar(m2)  
                   yield m1, m2, nqb    
    # Sparse states           
    for cols1 in range(2,6):
        for cols2 in range(2,6):
            for nqb in range(min(cols1, cols2)):
               for r in range(3):
                   for n in range(2):
                       m1 = qs_rand_matrix(0, cols1,r)
                       m1.mul_scalar(randint(-8, 8), randint(0,7))  
                       m2 = qs_rand_matrix(0, cols2, r)
                       m1.mul_scalar(randint(-8, 8), randint(0,7))  
                       yield m1, m2, nqb    
                


def qs_complex_prod(a, b, nqb, nc):
    """Complex equivalent to function ``qstate12_product``

    ``a`` and ``b`` must be complex numpy arrays of dimension 2.
    The function returns ``qstate12_product(a, b, nqb, nc)``
    as a complex umpy array of dimension 2.
    """
    nb = nqb - nc
    assert nb >= 0
    a1 = a.reshape((1 << nc, 1 << nb, -1))
    b1 = b.reshape((1 << nc, 1 << nb, -1))
    #prod = np.einsum("ikl,jkl->ijk", a1, b1)
    prod = np.einsum("mki,mkj->kij", a1, b1)
    prod = prod.reshape((-1,1))
    return prod


def check_eq_cols(qs1, qs2, nqb):
    """Check if the first ``nqb`` columns of states are equal
    
    Raise ValueError if this is not the case. 
    ``qs1`` and ``qs2`` must be instances of class ``QStateMatrix``.
    """
    s = 0
    n = min(qs1.nrows, qs2.nrows)
    m1, m2 = qs1.data, qs2.data
    cols1, cols2 = qs1.ncols - nqb, qs2.ncols - nqb
    for i in range(n):
        s |= (m1[i] >> cols1) ^  (m2[i] >> cols2)
    for i in range(n, qs1.nrows):
        s |= (m1[i] >> cols1) 
    for i in range(n, qs2.nrows):
        s |= (m2[i] >> cols2) 
    s &= (1 << nqb) - 1
    assert s == 0
   

@pytest.mark.qstate
def test_qs_prep_mul(verbose = 0):
    """Test function ``qstate12_prep_mul``. """
    for ntest, (m1, m2, nqb) in enumerate(create_product_testvectors()):
        _, m1a, m2a = prep_mul(m1, m2, nqb)
        if verbose:
            print("TEST %s" % (ntest+1))
            print(m1)
            print(m2)
            print("nqb =", nqb)
            print("Output states")
            print(m1a)
            print(m2a)
        check_eq_cols(m1a, m2a, nqb)
        c1 = qs_complex_prod(m1.complex(), m2.complex(), nqb, 0)
        c1a = qs_complex_prod(m1a.complex(), m2a.complex(), nqb, 0)
        try:
            err = "Function prep_mul() has failed"
            #compare_complex(c1, c1a, err)
        except ValueError:
            print("\n" + err +"\nInput states:")
            print(m1); print(m2);  
            print("nqb =", nqb)      
            print("\nOutput states:")
            print(m1a); print(m2a);  
            raise      


@pytest.mark.qstate
def test_qs_product(verbose = 0):
    """Test function ``qstate12_product``. """
    for ntest, (m1, m2, nqb) in enumerate(create_product_testvectors()):
        nc = randint(0, nqb)
        if verbose:
            print("TEST %s" % (ntest+1))
            print(m1)
            print(m2)
            print("nqb =", nqb, " nc =", nc)
            print("Output states")
        try:
            err = "Execution of function qstate12_product() has failed"
            m3 = flat_product(m1, m2, nqb, nc)
        except ValueError:
            print("\n" + err +"\nInput states:")
            print(m1); print(m2);  
            print("nqb =", nqb, ", nc =", nc)  
            row_pos, pm1, pm2 = qstate12_prep_mul(m1, m2, nqb)            
            print("prep_mul, row pos = ", row_pos) 
            print(QStateMatrix(pm1)); print(QStateMatrix(pm2))            
            print("\nOutput states:")
            qm1, qm2 = m1.copy(), m2.copy()
            qm1.qstate12_product(qm2, nqb, nc)
            print(m1); print(m2); 
            raise      
        c1, c2, c3 = m1.complex(), m2.complex(), m3.complex()
        c3_ref = qs_complex_prod(c1, c2, nqb, nc)
        try:
            err = "Function flat_product() has failed"
            compare_complex(c3_ref, c3, err)
        except ValueError:
            print("\n" + err +"\nInput states:")
            print(m1); print(m2);  
            print("nqb =", nqb, ", nc =", nc)      
            print("\nOutput states:")
            print(m3);  
            raise      
        if verbose:
            print("Output state")
            print(m3)
            
            
            
def create_matmul_testvectors():
    """Yield pairs of matrices for testing matrix multiplication

    These matrices are instances of class ``QStateMatrix``.
    """
    for i in range(2,6):
        for j in range(2,6):
            for k in range(2, 6):
                for r in range(3, 12, 2):
                    for n in range(1):
                        m1 = qs_rand_matrix(i, j, r)
                        rand_mul_scalar(m1)  
                        m2 = qs_rand_matrix(j, k, r)
                        rand_mul_scalar(m2)  
                        yield m1, m2   

@pytest.mark.qstate
def test_matmul(verbose = 0):
    """Test matrix multiplication. """
    for ntest,(m1, m2) in enumerate(create_matmul_testvectors()):
        if verbose:
            print("TEST %s QStateMatrix multiplication" % (ntest+1))
            print(m1)
            print(m2)
        m3 = m1 @ m2
        c1, c2 = m1.complex(), m2.complex()
        c3 = m3.complex()
        c3_ref = c1 @ c2        
        err = "Matrix multiplcation has failed"
        compare_complex(c3_ref, c3, err)
        if verbose:
            print("Product")
            print(m3)



def large_matmul_testvectors():
    """Yield test data for large matrix multiplication

    yield pairs ``m1``, ``m2`` of 2**12 times 2**12
    and also pairs random slices of about 40 elements
    in range(2**12)
    """
    nn = 12
    for r in range(10, 27):
        m1 = qs_rand_matrix(nn, nn, r)
        rand_mul_scalar(m1)  
        m2 = qs_rand_matrix(nn, nn, r)
        rand_mul_scalar(m2)
        r1 = slice(randint(0,100), 1<<nn, 2*randint(150,160)+1)       
        r2 = slice(randint(0,100), 1<<nn, 2*randint(150,160)+1)       
        yield m1, m2, r1, r2   


@pytest.mark.qstate
def test_large_matmul(verbose = 0):
    """Testlarge  matrix multiplication. 
    
    For the mmgroup project wee have to multiply marices
    of type QStateMatrix with shape (12,12). So we'd
    better test this operation here!
    
    We generate random matrices m1, m2. We test the 
    multiplication of a random  r times 4096 submatrix
    of m1 with a random 4096 times r submatrix of m2.
    Here r is about 14.
    """
    for ntest,(m1, m2, r1, r2) in enumerate(large_matmul_testvectors()):
        if verbose:
            print("TEST %s QStateMatrix multiplication" % (ntest+1))
            print(m1.echelon())
            print(m2.echelon())
        m3 = m1 @ m2
        if verbose:
            print("Product")
            print(m3)
        c1, c2 = m1[r1,:], m2[:,r2]
        c3 = m3[r1, r2]
        c3_ref = c1 @ c2        
        err = "Matrix multiplication has failed"
        compare_complex(c3_ref, c3, err)
        m3 = m1 @ m1
        c1, c2 = m1[r1,:], m1[:,r2]
        c3 = m3[r1, r2]
        c3_ref = c1 @ c2        
        err = "Matrix squaring has failed"
        compare_complex(c3_ref, c3, err)



def create_mul_testvectors():
    """Yield pairs of matrices for testing matrix multiplication

    These matrices are instances of class ``QStateMatrix``.
    """
    for i in range(2,6):
        for j in range(2,6):
            for r in range(2, 12, 2):
                for n in range(2):
                    m1 = qs_rand_matrix(i, j, r)
                    rand_mul_scalar(m1)  
                    m2 = qs_rand_matrix(i, j, r)
                    rand_mul_scalar(m2)  
                    yield m1, m2   



@pytest.mark.qstate
def test_mul(verbose = 0):
    """Test elementwise and scalar multiplication for `QStateMatrix``"""
    for ntest, (m1, m2) in enumerate(create_mul_testvectors()):
        if verbose:
            print("TEST %s Matrix multiplication of entries" % (ntest+1))
            print(m1.echelon())
            print(m2.echelon())
        m3 = (m1 * m2)
        if verbose:
            print("Product")
            print(m3)
        c3 = m3[:,:]
        c3_ref = m1[:,:] * m2[:,:]
        err = "Matrix multiplcation of entries has failed"
        compare_complex(c3_ref, c3, err)
        err = "Matrix negation has failed"
        compare_complex(-(m1[:,:]), (-m1)[:,:], err)
        err = "Matrix scalar multiplication failed"
        f = (1+1j)**randint(-8,7) * 2**(0.5*randint(-10,10))
        compare_complex(f*(m1[:,:]), (f*m1)[:,:], err)
        compare_complex(f*(m1[:,:]), (m1*f)[:,:], err)
        compare_complex((m1[:,:]/f), (m1/f)[:,:], err)
        if ntest < 20:
            compare_complex(0*(m1[:,:]), (0*m1)[:,:], err)
   

def create_conj_testvectors():
    """Yield pairs of matrices for testing matrix multiplication

    These matrices are instances of class ``QStateMatrix``.
    """
    for i in range(2,6):
        for j in range(2,6):
            for r in range(2, 12, 2):
                for n in range(2):
                    m1 = qs_rand_matrix(i, j, r)
                    rand_mul_scalar(m1)  
                    yield m1  

   
@pytest.mark.qstate
def test_conj(verbose = 0):
    """Test conjugation and transposition for `QStateMatrix``"""
    for ntest, m1 in enumerate(create_conj_testvectors()):
        if verbose:
            print("TEST %s Matrix multiplication of entries" % (ntest+1))
            print(m1.echelon())
        if verbose:
            print("Conjugate")
            print(m1.conj())
        err = "Matrix conjugation failed" 
        compare_complex(m1.conj()[:,:], m1[:,:].conj(), err)
        if verbose:
            print("Transpose")
            print(m1.T)
        err = "Matrix transpostion failed" 
        compare_complex(m1.T[:,:], m1[:,:].T, err)


     


from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

from random import randint
from collections import defaultdict

import pytest


from mmgroup.tests.test_mm_op.test_group_op import test_op, test_rand_op
from mmgroup.tests.test_mm_op.test_group_op import one_test_rand_op
from mmgroup.mm_space import characteristics
from mmgroup.tests.spaces.spaces import MMTestSpace
from mmgroup.structures.mm0_group import MM0


PRIMES = characteristics()

########################################################################
# Low level multiplication
########################################################################


""" Low level tests

The following function mul_group() is used as a low-level group
operation function in the test functions test_op() and test_rand_op().
This allows testing the main operations of the monster group on a 
lower level than the test in module test_group_op.
"""



def mul_atom(v1, tag, i, v2):
    data1 = v1.data
    data2 = v2.data
    mm = v1.ops
    if tag == 'd':
        mm.op_delta(data1, i, data2) 
    elif tag == 'p':
        mm.op_pi(data1, 0, i, data2)
    elif tag == 't':
        mm.op_t(data1,  i, data2)
    elif tag == 'l':
        mm.op_xi(data1,  i, data2)
    elif tag == 'x':
        mm.op_xy(data1, 0, i, 0, data2)
    elif tag == 'y':
        mm.op_xy(data1, i, 0, 0, data2)
    else:
        raise TypeError("Bad tag %s in monster operation" % str(t))
 

def mul_group(v, g):
    v = v.copy()
    v1, v2 = v, v * 0
    for tag, i in g.as_tuples():
        mul_atom(v1, tag, i, v2)
        v1, v2 = v2, v1
    if v != v1:
       v.data = v1.data[:]
    v.check()
    return v
   

@pytest.mark.mm_op
def test_low_group_op():
    print("")
    test_op(f_mul = mul_group, verbose = 0)
    test_rand_op(f_mul = mul_group, verbose = 0)



########################################################################
# Low level test for operation delta * pi
########################################################################

"Here we test the function op_pi() implementig a group operation."

def f_mul_delta_pi(v, g):
    t = g.as_tuples()
    d = defaultdict(int)
    d.update(t)
    v1  = v.copy()  # empty vector of same shape as v
    v.ops.op_pi(v.data, d['d'], d['p'], v1.data)
    return v1





@pytest.mark.mm_op
def test_op_delta_pi(verbose = 0):
    print("Testing group operation delta * pi")
    for p in PRIMES:
        space = MMTestSpace(p)
        group = space.group
        for i in range(50):
            g = MM0( [("d",'r'), ("p",'r')] )
            basis = "D" * 3 + "ABC" * 10 + "TXYZ" * 20 
            v = space('R')
            one_test_rand_op(v, g, basis, f_mul_delta_pi, verbose)
    print("passed")
            

########################################################################
# Low level test for operation y * x * delta
########################################################################

"Here we test the function op_xy() implementig a group operation."

def f_mul_yx(v, g):
    t = g.as_tuples()
    #print("ttt", dict(t))
    d = defaultdict(int)
    d.update(t)
    v1  = v.copy()  # empty vector of same shape as v
    v.ops.op_xy(v.data, d['y'], d['x'], d['d'], v1.data)
    return v1


@pytest.mark.mm_op
def test_op_yx(verbose = 0):
    print("Testing group operation yx")
    for p in PRIMES:
        space = MMTestSpace(p)
        group = space.group
        for i in range(50):
            g = group( [("y","r"), ("x","r"), ("d","r")] )
            basis = "D" * 3 + "ABC" * 10 + "TXYZ" * 20 
            v = space('R')
            one_test_rand_op(v, g, basis, f_mul_yx, verbose)
    print("passed")



def f_mul_omega(v, g):
    t = g.as_tuples()
    if len(t):
        assert t[0][0] == "x" and t[0][1] & ~0x1800 == 0
        x = t[0][1]
    else:
        x = 0
    v1 = v.copy()
    v.ops.op_omega(v1.data, x)
    return v1

@pytest.mark.mm_op
def test_op_omega(verbose = 0):
    print("Testing group operation x_Omega")
    for p in PRIMES:
        space = MMTestSpace(p)
        group = space.group
        for i in range(5):
            for d in (0, 0x800, 0x1000, 0x1800):
                g = group([("x", d)])
                basis = "D" * 3 + "ABC" * 10 + "TXYZ" * 20 
                v = space('R')
                one_test_rand_op(v, g, basis, f_mul_omega, verbose)
    print("passed")



"""Generate C code for Hadamard-like matrixes.

In [Seys19] we define two generators tau and xi of the monster
group and their representation matrices. Most of the non-monomial
parts of these matrices can be implemented as Hadamard matrices, 
which must be scaled by negative power of two. In order to obtain
tau and xi, the Hadamard matrices must also be multiplied by 
certain monomial matrices from the left and from the right.

Here a Hadamard matrix is a 2**n times 2**n matrix H with indices
running from 0 to 2**n - 1. Matrix H has entries 
H[i,j] = parity(i & j), with the '&' operator as in C and
parity(x) the bit parity of the nonnegative integer x, when x is 
given in binary notation.

This module contains a class HadamardMatrixCode for generating
C code that multiples a vector v with a Hadamard matrix H from 
the right. The entries of the vector v are integers modulo p for 
a fixed (not necessarily prime) number p = 2**k-1, 2 <= k <= 8.
 
Vector v is coded as an array of integers of type uint_mmv_t,
where uint_mmv_t is uint32_t or uint64_t, depending on the
settings in module config.py. Here an single integer of type 
uint_mmv_t may store several components of vector v as described
in module mm/mm_aux.py. Class HadamardMatrixCode is a subclass
of class mm/mm_basics.MM_Op, which contains basic operations for
dealing with such a vector.

Our hope is that an optimzing C compiler completely performs the 
relevant operations in registers, if sufficiently many of them 
are present. Note that fewer operations are necessary if v consists 
of 64-bit integers. On the other hand, modelling 64-bit integers on 
a 32-bit architecture causes unnecessary overhead. So we support 
both, 32-bit and 64-bit integers.  



References
----------
see file references.txt
"""



#23456789012345678901234567890123456789012345678901234567890123456789012345

from __future__ import absolute_import, division, print_function
#from __future__ import  unicode_literals



import sys
import os
from numbers import Integral

from mmgroup.bitfunctions import bitparity, bitweight
from mmgroup.dev.mm_op.mm_op import MM_Op, INT_BITS, c_hex

####################################################################
FAST_MOD3 = True # Use new fast code for P = 3 when set

def as_c_expr(x):
    """Map integers to hex string, don't change anything else

    This function is used for entering objects into automatically
    generated C programs in the following class C_Expr, which 
    models expressions in the C language.  

    Most of these objects have a 'str' or '__repr__' method that
    formats the object as required for the C language.

    This would also work for integer constants, when writing 
    them as decimal numbers. In the following class 
    HadamardMatrixCode almost all integer constants are bit
    masks, so we better write them as hexadecimal constants
    of sufficient bit length. Function c_hex in module mm_basics 
    does this job.
    """
    if isinstance(x, Integral):
        return c_hex(x)
    return x



def n_ops(x):
    return x.n_ops if isinstance(x, C_Expr) else 0

class C_Expr:
    """Models an arbitrary C expression

    Applying a standard operator such as '+', '-', '&', '|', '^', 
    '>>', '<<',  to an instance of this class returns the C code 
    for performing the corresponding operation as a string. 

    Atomic C expressions, e.g. variables, must be given as instan-
    ces of subclasses of this class; see e.g. class C_UintVar.
    One operand of an operation with an instance of this class
    may be an integer. Then the corresponding integer constant
    (written hexadecimally) is used as that operand. Note that
    shift factors in the '<<' and '>>' operators must be integers.

    Here only the C operators required for our purposes are 
    implemented. We do not make any attempt to save brackets.
    """
    def __init__(self, name, n_ops = None):
        self.name = name
        self.n_ops = 0 if n_ops is None else n_ops

    def __repr__(self):
        return self.name

    str = __repr__

    def __lshift__(self, other):
         assert isinstance(other,int), "2nd shift operand must be integer"
         if other == 0:
              return self
         elif other > 0:
              return C_Expr("(%s << %d)" % (self, other), self.n_ops + 1) 
         elif other < 0:
              return C_Expr("(%s >> %d)" % (self, -other), self.n_ops + 1) 

    def __rshift__(self, other):
         return self.__lshift__(-other)

    def _binop(self, other, op):
         oper1 = str(as_c_expr(self))
         oper2 = str(as_c_expr(other)).strip()
         ops = self.n_ops + n_ops(other) + 1
         if len(oper1) + len(oper2) > 40:
             op_ = "\n    " + op + " "
         else:
             op_ = " " + op + " "
         return  C_Expr("(" + oper1 + op_ + oper2 + ")", ops) 

    def __add__(self, other):
        return self._binop(other, '+')

    __radd__ = __add__

    def __sub__(self, other):
        return self._binop(other, '-')

    def __and__(self, other):
        return self._binop(other, '&')

    __rand__ = __and__

    def __or__(self, other):
        return self._binop(other, '|')

    __ror__ = __or__

    def __xor__(self, other):
        return self._binop(other, '^')

    __rxor__ = __xor__

    def __invert__(self):
        return  C_Expr("(~%s)" % self)

    def __neg__(self):
        return  C_Expr("(-%s)" % self)

    def assign(self, value):
        """Return string encoding an assignment statement: self = value;
     
        'value' must be another instance of this class or a string  
        containing a valid C expression or an integer.
        """ 
        return  "%s = %s;\n" % (self, as_c_expr(value))
 
    def assign_xor(self, value):
        """Return string encoding an assignment statement: self ^= value;
     
        'value' must be another instance of this class or a string  
        containing a valid C expression or an integer.
        """ 
        return "%s ^= %s;\n" % (self, as_c_expr(value))

    def assign_and(self, value):
        """Returnstring encoding an assignment statement: self &= value;
     
        'value' must be another instance of this class or a string  
        containing a valid C expression or an integer.
        """ 
        return "%s &= %s;\n" % (self, as_c_expr(value))

    def assign_or(self, value):
        """Return string encoding an assignment statement: self |= value;
     
        'value' must be another instance of this class or a string  
        containing a valid C expression or an integer.
        """ 
        return "%s |= %s;\n" % (self, as_c_expr(value))


class C_UintVar(C_Expr):
    """Models a special integer variable of a given type

    Instances of this class model integer variables which are designed 
    for use with C operations as specified in class C_Expr.

    Here a variable has a name and a type, which must be a valid
    integer type in the C code to be generated.

    Such variables are also designed to be a member of a list of 
    variables. Each variable as an optional attribute 'index'
    specifying the index of the varible in that list.
    """
    def __init__(self, c_type, name, index = None, var_number = None):
         """Create a C variable with a given type and 'name'.

         The type of the variable is the string given in the
         argument 'c_type'. 'index' is an optional argument
         specifiying the index of the variable in a list.
         This list is kept by the code generating  process 
         only; and it is irrelevant for the C compiler.

         Usually variables are created as instances of this
         class when needed. Then their names contain
         consecutive numbers, e.g. 'r0', 'r1', 'r2', ...    
         The corresponding number of a variable may be 
         passed with argument 'var_number'.
         """
         self.name = name
         self.n_ops = 0
         self.var_number = var_number
         self.index = index
         self.c_type = c_type 
         self.context = None
         self.excess = 0


    def set_context(self, context):
        """Set a context for a variable

        Usually a context is an instance of class ``HadamardMatrixCode``.
        If a context ``ct`` has been set than method

        self.assign(value)

        is equivalent to

        self.context.assign(self, value)

        Similarly,

        self.assign_xxx(value)

        is equivalent to        

        self.context.assign_xxx(self, value)

        for 'xxx' = 'and', 'or', 'xor'.
        """
        self.context = context

    def assign(self, value):
        """Generate C code computing 'self = value;' in a context

        When calling the method, a conext must have been set with 
        method ``self.set_context``.

        Similarly,

        self.assign_and(value)  generates code 'self &= value;'

        self.assign_or(value)  generates code 'self |= value;'

        self.assign_xor(value)  generates code 'self ^= value;'
        """
        self.context.assign(self, value)

    def assign_and(self, value):
        self.context.assign_and(self, value)
    def assign_or(self, value):
        self.context.assign_or(self, value)
    def assign_xor(self, value):
        self.context.assign_xor(self, value)


class C_UintVarPool:
    """Model a (usually short) list of integer variables.

    Assume that we want to modify a short vector v of integers 
    stored in memory. Therefore we load all entries of vector v
    into variables of the appropriate type. Then we perform the
    requested operations on these variables. Finally, we store
    all variables back to memory. Thus an optimizing compiler
    may save run time by keeping some or all variables in 
    registers. 

    So we may create a pool of variables of a given type with
    a given name. Here name = "r" means that the variables of
    the pool with have names "r0", "r1", "r2",... .

    All variables in the pool are instances of class C_Expr,
    so we may perfor C operations on them as decribed in 
    class C_Expr.

    The pool has an 'official'  size equal to len(self),
    which is set in the constructor, and which can be modified
    by method self.extend_pool(new_size). The official
    variables are ordered, so self[i] is the  i-th official
    variable in the pool. This mean that the official variables 
    may represent a vector of integers. The order of the
    variables can be modified with method self.xch(); so
    exchanging components of the vector is for free.

    Apart from the official variables there are temporary
    variables of the same type in the pool. self.temp(i) 
    returns the i-th temporary variable. We often need 
    temporary variables for the operations on the entries of 
    the vector. Official variables can also be exchanged
    with temporary variables.

    An instance of this class keeps track of all variables 
    created. Method self.declare() can be used to declare all 
    these variables in a C program.
    """
    LINE_LEN = 50
    context = None

    def __init__(self, c_type = "uint_mmv_t", name = "r%d", length = 0):
        """Create a pool of variables of type 'ctype'.

        The pool will contain 'length' official variables. These
        variables are instances of class C_UintVar and they are
        obtained as self[0], ..., self[length - 1].

        The names of the variables are name % 0, name % 1, name % 2, ..
        where name is given by parameter 'name'. So for obtaining
        r0, r1, r2, ... parameter 'name' should be "r%d"
        """
        self.a_size = 0      # Number of official variables
        self.vars = []       # List of all veriables, Here 
                             # official variable are listed
                             # in their current order, followed
                             # by the temporary variables.
        self.name = name     # Name string used for variables
        self.c_type = c_type # C type of the variables
        self.resize(length)  # Create 'length' official variables
         

    def set_context(self, context):
        """Set a context for the variables in this pool.

        This means that we will apply method ``x.set_context(context)``
        to any new variable ``x`` in htis pool
        """
        self.context = context
        for v in self.vars:
            v.set_context(context)


    def extend_pool(self, new_size):
        """Internal method

        The method adds variables to the pool so that there are
        at least 'new_size' variables in the pool.
        """
        while len(self.vars) < new_size:
            l = len(self.vars)
            name = self.name % l
            var = C_UintVar(self.c_type, name, l, l)
            if self.context:
                var.set_context(self.context)
            self.vars.append(var)

    def __getitem__(self, i):
        """Return the i-th official variable"""
        assert 0 <= i < self.a_size, (i, self.a_size)
        return self.vars[i]

    def resize(self, new_size):
        """Change the size of the pool to 'new_size'

        Afterwards, the 'offical' part of the pool has has the
        length given by 'new_size'. All temporary variables in
        the pool are discarded.
        """
        self.extend_pool(new_size)
        self.a_size = new_size

    def temp(self, i = 0):
        """Return the i-th temporary variable"""
        self.extend_pool(self.a_size + i + 1)
        return self.vars[self.a_size + i]

    def xch(self, i, j):
        """Exchange self[i] with self[j].

        i and j may also be variables in the pool (including
        temporary variables) insted of their indices. Thus

           v1 = self[1]; t = self.temp(0); self.xch(v1, t)

        exchanges an official variable with a temporary variable.

        This method performs just book keeping, so that we need 
        not generate any code here.   
        """
        if isinstance(i, C_UintVar):
            var_i, i = i, i.index 
        else:
            var_i = self.vars[i]
        if isinstance(j, C_UintVar):
            var_j, j = j, j.index 
        else:
            var_j = self.vars[j]
        assert self.vars[i].index == i
        assert self.vars[j].index == j
        if i == j:
            return
        assert 0 <= i < len(self.vars)
        assert 0 <= j < len(self.vars)
        self.vars[i], self.vars[j] = self.vars[j], self.vars[i]
        self.vars[i].index = i
        self.vars[j].index = j

    def declare(self):
        """Declare all variables in the pool in C

        The function returns a string containing that declaration.
        """
        line_len = self.LINE_LEN - len(self.c_type) - 1
        nvars =  line_len // (len(self.name) + 4)
        pos = 0
        s = ""
        while pos < len(self.vars):
           end_pos = min(pos + nvars, len(self.vars))
           vars = [self.name %i for i in range(pos, end_pos)]
           s += self.c_type + " " + ", ".join(vars) + ";\n"
           pos = end_pos
        return s
         
    def __iter__(self):
        """Iterate over official variables in their current order"""
        for var in self.vars[:self.a_size]:
            yield var 

    def __len__(self):
        """Return the number of the official variables"""
        return self.a_size


    def var_number_list(self):
        """Return list of numbers of the official variables

        Given a 'name' for the variables of the pool, the
        official variables have names <name>0, <name>1, <name>2 ..
        These names may be permuted by method self.xch().

        This method returns the list of numbers in the names
        of the official variables in the current order. 
        """
        return [self.vars[i].var_number for i in range(self.a_size)]


class HadamardMatrixCode(MM_Op):
    """Generate C code for multiplying vector by a Hadamard matrix 
 
    We generate C code for multiplying a vector v of integers
    modulo p = 2**k-1  by a Hadamard matrix H. 

    Vector v is stored in memory as described in class
    mm/mm_aux.py. It is loaded into a set of C variables
    which is constructed as an instance of class 
    C_UintVarPool. C code for multiplication of v by H is 
    generated with the operations given in class C_Expr,
    operating on these variables. Finally the result is
    written back from the variables to memory.

    Matehmatically, a 2**n times 2**n times Hadamard matrix
    is constucted as product of n matrices of shape

       I2  (x) ... (x) I2 (x) H2 (x) (I2) (x) ... (x) (I2) , 
    
    where (x) denotes the Kronecker product, I2 is the 2 times 2 
    unit matrix, and H2 is the 2 times 2 Hadamard matrix. H2 
    occurs exactly once in each of the k factors of the matrix
    product. In different factors, H2 occures at different 
    positions. Note that all these k matrix factors commute. 

    The m-th matrix factor of H is implemented as a butterfly 
    operation on vector v in the usual way:

         v[i], v[i+j] = v[i] + v[i+j], v[i] - v[i+j],

    with j = 2**m, for all integers i where bit m of i is zero. 
    """
    def __init__(self, p, log_vlength, verbose = 0):
        """Create an instance of class HadamardMatrixCode

        Operation on vectors are done modulo p as described
        in class mm/mm_aux.py. p must be of shape
        p = 2**n-1, 2 <= n <= 8.

        'log_vlength' is the binary logarithm of the size of
        the vector v and of the Hadamard matrix H. We test
        the cases 'log_vlength' = 2, 4, 6 corresponding to
        vectors of size 4, 16, 64 only.
        """
        super(HadamardMatrixCode, self).__init__(p)
        self.LOG_VLEN = log_vlength   
        self.NO_CARRY = self.FIELD_BITS == self.P_BITS
        self.FAST_MOD3 = FAST_MOD3 and self.P == 3
        self.NO_CARRY &= not self.FAST_MOD3
        self.verbose = verbose
        self.make_h_masks()
        self.reset_vars()
        if verbose:
           print("HadamardMatrixCode mod", p,  ", size =", self.vlen)


    def reset_vars(self, log_vlen = None):
        """Reset variables 

        This method resets all information required for the 
        the code generation process, including the pool self.vars 
        of variables, which is an intance of class C_UintVarPool.

        If log_vlen is specified, self.LOG_VLEN is changed to log_vlen.        
        """
        if log_vlen: self.LOG_VLEN = log_vlen
        if self.verbose: print("reset vars")
        self.vlen = max(1, (1 << self.LOG_VLEN) >> self.LOG_INT_FIELDS)
        self.vars = C_UintVarPool("uint_mmv_t", "r%d", self.vlen)
        self.vars.set_context(self)
        self.expanded = 0
        self.matrix_code = []     # C implementation of matrix operation
        self.n_code_lines = 0     # just for staticstics
        self.n_operations = 0     # just for staticstics
        self.reset_butterfly()
      

    def make_h_masks(self):
        """Create bit masks required for a butterfly operation

        self.ALL1_MASK is the all-bits-one mask.

        For a power i of two, self.B0_MASK[i] selects the 
        entries v[j] of a vector stored in a single variable 
        with i & j = 0, and self.B1_MASK[i] selects the 
        entries v[j] with i & j = i.

        self.H_MASK_EXTERN is for butterfly op. on different 
        variables var[i], var[j]. It selects all valid fields 
        of a variable. self.H_MASK_CARRY selects the carry
        bits of all valid fields

        self.H_MASK[i] is for butterfly op. on (v[j], v[j+i]),
        with v[i] nd v[i+j] stored in the same variable, for 
        the relevant powers i of two. Here the mask selects 
        the components v[j] of v with i & j == i.

        Note that in case p = 2**2**k-1, (i.e.  p = 3, 15, 255),
        the bit fields with odd index in a variable are invalid
        when a vector is expanded for a Hadamard operation, as
        explained in method expand_hadamard(). self.H_MASK[j]
        takes care of this special situation.
        """
        self.ALL1_MASK = (1 << self.INT_BITS) - 1
        self.B0_MASK = {} 
        self.B1_MASK = {} 
        self.H_MASK = {} 
        NC = bool(self.NO_CARRY)
        self.H_BITS = H_BITS = self.FIELD_BITS << NC
        self.H_MASK_EXTERN = self.smask(self.P, -1, H_BITS) 
        self.H_MASK_CARRY =  self.smask(self.P+1, -1, H_BITS)
        for i0 in range(self.LOG_INT_FIELDS):
            i = 1 << i0
            mask = [x for x in range(self.INT_FIELDS) if x & i == 0]
            self.B0_MASK[i] = self.smask(self.P, mask)
            mask = [x for x in range(self.INT_FIELDS) if x & i == i]
            self.B1_MASK[i] = self.smask(self.P, mask)
            self.H_MASK[i << NC] = self.smask(-1, mask, H_BITS)
        if self.P == 3:
            self.BIT1_MASK = self.smask(2, -1, self.FIELD_BITS)
       
    def add(self, s, n_lines = 0, n_ops = 0):
        """Add the string ``s`` to the generated C code.

        String ``s`` must contain a linefeed character if such
        a character is to be addedto the C code.

        The line counter and the operation counter are incremented 
        by the optional parameters ``n_lines`` and ``n_ops``.
        """
        self.matrix_code.append(s)
        self.n_code_lines += n_lines
        self.n_operations += n_ops

    def __iadd__(self, s):
        """``self += s`` is equivalent to ``self.add(s)``"""
        self.matrix_code.append(s)
        return self
      

    def comment(self, comment):
        """Enter a 'comment' into the C code"""
        for s in comment.split("\n"):
            self.matrix_code.append("// " + s + "\n")

    def comment_vector(self):     
        """Comment of represention of vector v in  variables"""
        vl = self.vars.var_number_list()
        vdata = ",".join(map(str,vl)) 
        if len(vl) > 10: vdata = "\n " + vdata 
        name = self.vars.name.replace("%d", "%s") % "(i)"        
        self.comment("Vector is now  %s for i = %s" % (
            name, vdata))

    def comment_statistics(self, reset = True):
        """Write statistics as comment into the C code"""
        self.comment("%d lines of code, %d operations" %
               (self.n_code_lines, self.n_operations))
        if reset:
            self.n_code_lines = self.n_operations = 0;

    def _assign(self, var, value, op = ""):
        ops = n_ops(value) + bool(op)
        s_ops = " // %d ops" % ops if ops > 1 else ""
        s =  "%s %s= %s;%s\n" % (var, op, as_c_expr(value), s_ops)
        self.add(s, 1, ops)

    def assign(self, var, value):
        """Generate C code computing 'var = value;'"""
        self._assign(var, value)

    def assign_and(self, var, value):
        """Generate C code computing 'var &= value;'"""
        self._assign(var, value, '&')
      
    def assign_or(self, var, value):
        """Generate C code computing 'var |= value;'"""
        self._assign(var, value, '|')
      
    def assign_xor(self, var, value):
        """Generate C code computing 'var ^= value;'"""
        self._assign(var, value, '^')
      

    def reset_butterfly(self, lazy = False):
        """Reset attribute for butterfly operation.

        A Hadamard operation is essentially a sequence of butterfly
        operations ``(x[i], x[j]) = (x[i] + x[j], x[i] - x[j])``
        modulo p. 

        If parameter ``lazy`` is ``False`` (default) then  we will 
        reduce all entries ``x`` of a vector modulo p immediately 
        after an addition or subtraction  such such that 
        ``0 <= x <= p`` holds. Therefore we use 
        method ``reduce_final``.

        If parameter ``lazy`` is ``True`` then we will reduce these
        entries only before an overflow (across a bit field of
        width ``self.H_BITS``) may occur. Therefore we use 
        method ``prereduce`` when necessary and method 
        ``reduce_final`` once at the end of a Hadamard operation.

        Setting  ``lazy = True`` may lead to more efficient code if 
        self.H_BITS - self.P_BITS > 3, say. The function stores
        parameter ``lazy`` in the attribute ``self.lazy``. It may
        put ``self.lazy = 0`` if lazy reduction is impossible or
        inefficient.
        """
        self.FREE_BITS = self.H_BITS - self.P_BITS
        self.H_MASK_EXTERN_HI = ((self.H_MASK_EXTERN ^ self.ALL1_MASK) 
             >> self.P_BITS)
        self.lazy = lazy and self.FREE_BITS > 2 and not self.FAST_MOD3
        self.shift_stage = 0
        for v in self.vars:
            v.excess = 0


    def add_mod3(self, var_a, var_b, var_c, var_tmp = None):
        """Put var_c = var_a + var_b (mod 3).
   
        This function performs the complete addition modulo 3,
        including reduction mod 3. This works for self.p == 3
        and self.P_BITS == 2 only. It uses a temporary variable
        var_temp.
        """
        assert self.P == 3 and self.P_BITS == 2
        t = self.vars.temp() if var_tmp is None else var_tmp
        t.assign(var_a & var_b);
        t.assign_or((t << 1) & (var_a ^ var_b))
        t.assign_and(self.BIT1_MASK)
        var_c.assign(var_a + var_b - t - (t >> 1))
       
     
    def reduce_final(self, var, dest = None):
        """Auxilary reduction function for method butterfly_op().
   
        Performs reduction mod p after a butterfly operation for all 
        components ``var[i]`` stored in variable ``var``. The function 
        requires ``0 <= var[i] <= 2*p`` for all entries of the variable.
        After reduction we have ``0 <= var[i] <= p`` for all  entries.
        """
        if dest is None:
            dest = var
        if self.FAST_MOD3:
            if dest != var:
                dest.assign(var)
            return 
        t = self.vars.temp() if var == dest else dest
        t.assign(var & self.H_MASK_CARRY)
        dest.assign(var - t + (t >> self.P_BITS)) 

    def prereduce(self, var, dest = None):
        """Auxilary reduction function for method butterfly_op().
   
        Performs reduction mod p after a butterfly operation for all 
        components ``var[i]`` stored in variable ``var``. The function 
        requires ``0 <= var[i] < 2**self.H_BITS`` for all entries of 
        the variable. After reduction we have ``0 <= var[i] <= 2*p`` 
        for all  entries.

        Thif function is used in case ``self.lazy = True`` only.
        """
        if dest is None:
            dest = var
        if var.excess:
            var.assign(var - var.excess)
        dest.assign((var & self.H_MASK_EXTERN) +
            ((var >> self.P_BITS) & self.H_MASK_EXTERN_HI))
        dest.excess = 0


    def reduce_butterfly(self, var, dest = None):
        """Auxilary reduction function for method butterfly_op().

        This function sould be called after each bufferfly
        operation for all participating variables ``var``.
        The reduced C veriable ``var`` is stroed in the C
        variable ``dest``.
        """
        if self.lazy:
            if self.shift_stage >= self.FREE_BITS - 1:
                self.prereduce(var, dest)
            elif not dest in [var, None]:
                dest.assign(var)
        else:
            self.reduce_final(var, dest)

    def reduce_butterfly_final_all(self):
        """Finaly reduction function for method butterfly_op().

        This function should be called once after completing a
        Hadamard operation. At exit we have ``0 <= x <= p`` 
        for all entrie participating on the Hadamard operation.
        """
        if self.lazy:
            self.comment("Final reduction")
            for var in self.vars:
                if self.shift_stage > 1:
                    self.prereduce(var)
                self.reduce_final(var)
        self.shift_stage = 0

    def increment_shift_stage_after_butterfly(self):
        """Adjust size of entries after butterfly operations

        This method is relevant in case ``self.lazy = 1``only.
        In this case the lengths of the entries grow by one bit
        after completing a sequence of butterfly operations at the 
        same level. A butterfly operation reduces the size of an 
        entry before overflow can occur.

        This function sould be called after completing a sequence 
        of butterfly operations at the same level, in order to
        adjust the member ``self.shift_stage`` counting the
        excess bits of an entry.
        """
        if self.lazy:
            self.shift_stage += 1
            if self.shift_stage == self.FREE_BITS:
                self.shift_stage = 1
          
                


    def negate_expression(self, var, j = None):
        """Return expression negating certain fields of variable ``var``

        More_precisely the function returns a pair ``(expr, excess)``,
        where ``expr`` is C expression and ``excess`` is an integer
        constant such that ``vneg = expr - excess`` is the computed
        expression that negates some fields of ``var``.

        Here ``var`` must be a C variable and ``j`` must be a power of
        two. Then the computed expression ``vneg`` contains the entries
        of variable ``var``, where the entries of ``var`` with 
        indices ``i`` are negated in case ``i & j != 0`.

        If ``j`` is ``None`` (default) then the all  entries of ``var``
        are negated.
        """
        msk = self.H_MASK_EXTERN 
        if not j is None:
             msk &= self.H_MASK[j] 
        if self.shift_stage == 0:
            return var ^ msk, 0
        shifted_msk = msk << self.shift_stage
        xor_msk = msk | shifted_msk
        excess = (var.excess ^ xor_msk) - shifted_msk
        return var ^ xor_msk, excess
           

    def internal_butterfly(self, var, j):
        """Auxilary function for method butterfly_op().
   
        Perform butterfly operation inside the single C variable 
        'var'.  For all components v[i] of v stored in var we put

            v[i], v[i+j] = v[i] + v[j], v[i] - v[j] , 

        for j a power of two, j < self.INT_FIELDS, if the bit
        of valence j in i is equal to 0.

        Each result is reduced modulo p.
        """
        bsh = j << self.LOG_FIELD_BITS
        msk = self.H_MASK[j] 
        t = self.vars.temp() 
        if 2 * bsh  == self.INT_BITS:
            # We may save a bit in this case
            self.assign(t, (var << bsh) | (var >> bsh))
        else:
            self.assign(t, ((var << bsh) & msk) | ((var & msk) >> bsh))
        neg_var, neg_excess = self.negate_expression(var, j)
        if self.FAST_MOD3:
            t1 = self.vars.temp(1) 
            var.assign(neg_var)
            self.add_mod3(var, t, var, t1)
        else: 
            var.assign(neg_var + t)
            self.reduce_butterfly(var)
            if self.lazy:
                 x = var.excess
                 x = ((x << bsh) & msk) | ((x & msk) >> bsh)
                 var.excess = x + neg_excess


    def external_butterfly(self, v1, v2):
        """Auxilary function for method butterfly_op().
   
        Perform butterfly operation on the two C variables v1
        and v2.  For all components v1[i], v2[i] of of v1, v2
        (stored in v1 and v2) we put

            v1[i], v2[i] = v1[i] + v2[i], v1[i] - v2[i] , 

        Each result is reduced modulo p.
        """
        neg_v2, neg_excess = self.negate_expression(v2)
        if self.FAST_MOD3:
            t1, t2 = self.vars.temp(0), self.vars.temp(1) 
            self.add_mod3(v1, v2, t1, t2)
            v2.assign(neg_v2)
            self.add_mod3(v1, v2, v2, t2)
            v1.assign(t1)
        else:
            t = self.vars.temp()
            t.assign(v1 + neg_v2)
            v1.assign(v1 + v2)
            if self.lazy:
                x1, x2 = v1.excess, v2.excess
                v1.excess = x1 + x2
                t.excess = v2.excess = x1 + neg_excess 
            self.reduce_butterfly(t, v2)
            self.reduce_butterfly(v1)

    def external_butterfly_all(self, j):
        """Auxilary function for method butterfly_op().
   
        Performs a complete butterfly operation, where each single
        operation comprises two C variables.

        For all components v[i] of v  we put

            v[i], v[i+j] = v[i] + v[j], v[i] - v[j] , 

        of j a power of two, j >= self.INT_FIELDS, if the bit
        oc valence j in i is equal to 0.
        """
        if self.verbose: 
            print("ext butterfly", j, self.INT_FIELDS)
        assert j >= self.INT_FIELDS
        y = j >> self.LOG_INT_FIELDS
        for i in  range(len(self.vars)):
            if i & y == 0:
                self.external_butterfly(self.vars[i], self.vars[i + y])

    def lazy_shift(self, shift):
        """Try to do a shift data within a butterfly_operation

        The function tries to do multiply the vector by the scalar
        ``2**shift`` if the is can be done easily within a buterfly
        operation.

        It returns True if the code for this operation has been
        generated and False otherwise. In case of failure the same
        operation can be done with method ``mul_pwr2``.
        """
        shift1 = shift % self.P_BITS
        if shift1 == 0:
            s = "Multiplication by 2**%d is trivial mod %d"
            self.comment(s % (shift, self.P))
            return True
        if not self.lazy:
            return False
        doable = 1 <= shift1 <= self.FREE_BITS - self.shift_stage
        doable &= self.shift_stage > 1
        if not doable:
            return False
        s = "Multiply vector by scalar 2**%d mod %d"
        self.comment(s % (shift, self.P))
        for var in self.vars:
            var.assign(var << shift1)
            var.excess <<= shift1
        self.shift_stage += shift1
        return True



    def hadamard_op(self, operations = -1, shift = 0, compress = False):
         """Multiply vector v by Hadamard matrix H.

         Internal operation:

         For sh = 0, ..., self.LOG_VLEN - 1 we do the following:
         Let y = 1 << sh. Then we put

           v[i], v[i+y]  = (v[i] + v[y]) % p , (v[i] - v[i+y]) % p

         for all i with i & y == 0.

         The result is multipled with the fixed scalar ``2**shift`` 
         (mod p). Here ``shift`` defaults to 0.

         Remerk on internal operation:

         In cases p = 3, 15, 255, perform method expand_hadamard()
         before doing any butterfly operations. After that 
         expansion, sh has to run through  1,...,self.LOG_VLEN 
         instead. This expansion is reversed after the operation,
         so that the user need not bother about this special case. 

         If parameter ``compress`` is True then compression is
         always done, even if expnasion has bee done befor calling
         this function.
         """
         operations &= (1 << self.LOG_VLEN) - 1
         pre_expanded = self.expanded
         if not self.expanded:
             self.expand_hadamard(self.LOG_VLEN)
         expand_ = (operations & 1) << self.expanded
         operations &=  ~1 & ~(1 << self.expanded)
         operations |=  expand_
         lazy = self.P == 15 and bitweight(operations >= 6)
         self.reset_butterfly(lazy)
         if self.verbose:
             s = "Butterfly p=%d, vars=%d, op=%s, sh=%d, lazy=%d, width=%d, free=%d"
             print(s % (self.P, len(self.vars), hex(operations), shift,
                self.lazy, self.H_BITS, self.FREE_BITS)
            )     
         for lg_i in range(self.LOG_VLEN + 1):
             i = 1 << lg_i
             if i & operations == 0:
                 continue
             cmt = "Butterfly: v[i], v[i+%d] = v[i]+v[i+%d], v[i]-v[i+%d]" 
             self.comment(cmt  % (i, i, i))
             if i < self.INT_FIELDS:
                 for var in self.vars:
                     self.internal_butterfly(var, i)     
             else:    
                 self.external_butterfly_all(i)
             self.comment_vector()
             self.increment_shift_stage_after_butterfly()
         shift_done = self.lazy_shift(shift)
         self.reduce_butterfly_final_all()
         if not pre_expanded or compress:
             self.compress_hadamard()
         if not shift_done:
             self.mul_pwr2(shift)
         self.reset_butterfly()


    def expand_hadamard_intern(self, v1, position):
        """Auxiliary method method for method expand_hadamard()

        It operates on a variables v1 type uint_mmv_t.
        It zeros the components v1[2*i + 1] in vector v1 and it
        copies the old component v1[2*i + 1] to component 
        v1[2*i + 2**position].
        Components v1[2*i + 1] of v1 are set to zero.
        """
        assert 1 <= position < self.LOG_INT_FIELDS
        SH = (self.FIELD_BITS << position) - self.FIELD_BITS
        LO_MASK = self.B0_MASK[1] & self.B0_MASK[1 << position] 
        HI_MASK = LO_MASK << self.FIELD_BITS
        self.assign(v1, (v1 & LO_MASK) | ((v1 & HI_MASK) << SH))

    def expand_hadamard_intern_all(self, position):
        """Auxiliary method method for method expand_hadamard()

        It executes method expand_hadamard() in case 
        1 <= position < self.LOG_INT_FIELDS.
        """
        for var in self.vars:
            self.expand_hadamard_intern(var, position)


    def expand_hadamard_extern(self, v1, v2):
        """Auxiliary method method for method expand_hadamard()

        It operats on two variables v1, v2 of type uint_mmv_t.
        It zeros the components v1[2*i + 1] in vector v1 and it
        copies the old component v1[2*i + 1] to component v2[2*i].
        Components v1[2*i + 1] of v1 are set to zero.
        """
        mask = self.H_MASK_EXTERN
        self.assign(v2, (v1 >> self.FIELD_BITS) & mask )
        self.assign(v1, v1 & mask)


    def expand_hadamard_extern_all(self, position):
        """Auxiliary method method for method expand_hadamard()

        It executes method expand_hadamard() in case 
        position >= self.LOG_INT_FIELDS.
        """
        assert self.LOG_INT_FIELDS <= position <= self.LOG_VLEN
        if position == self.LOG_VLEN:
            self.vlen <<= 1
            self.vars.resize(self.vlen)
        s = ""
        j = 1 << (position - self.LOG_INT_FIELDS)
        mask = self.H_MASK_EXTERN
        for i in range(self.vlen):
            if i & j == 0:
                v1, v2 = self.vars[i], self.vars[i + j]
                self.expand_hadamard_extern(v1, v2)
                

    def expand_hadamard(self, position = -1):
        """This is a very peculiar function for p = 2**2**j-1. 

        Then each component v[i] of v is 2**j bits long. E.g. in case 
        p = 3 we have j = 1 and v[i] is 2 bits long. 
        For a butterfly operation we need one more bit to store the carry
        bit after an addition or subtraction of two components v[i], v[j].
        Thus a butterfly operation cannot be done immediately after 
        loading a vector v into C variables.
                
        This function splits one C variable into two parts, so that 2*2**j
        bits are available for each component v[i], and hence butterfly 
        operations can be done. Use method compress_characteristic_p()
        to undo the effect of this function before storing the vector v to
        to an array.
        
        A slightly nasty side effect is that this function changes the
        order of the components of v as follows:

        Let l be the length of vector v. After calling this method 
        the components of v are ordered a follows:

           0, 2, ..., l-2,   1, 3, ..., l-1 .

        At present this method is used by method hadamard_op()
        only. Method hadamard_op() takes care of this change
        and it restores the vector to its original form.
        So at present the user need not bother about this case.

        By default, entry  2*i+1  is moved to position
        2*i + 2**self.LOG_VLEN.  If the optional argument
        'position' satisfies 1 <= position < self.LOG_VLEN
        then entry  2*i+1  is moved to position
        2*i + 2**position  instead, overwriting the enty at
        thas position. This may be useful in cases where 
        the entries at those positions are not nneded.
        """ 
        if self.NO_CARRY and not self.expanded:
            if not 0 < position < self.LOG_VLEN: 
                 position = self.LOG_VLEN
            if self.verbose:   
               print("expand vector for Hadamard operation")          
            self.comment(
            """Expansion for Hadamard operation:
There is no space for a carry bit between bit fields. So 
we move bit field 2*i + 1  to bit field 2*i + %d.""" % (
                  1 << position)
            )
            if position < self.LOG_INT_FIELDS:             
                self.expand_hadamard_intern_all(position)
            else:
                self.expand_hadamard_extern_all(position)
            self.comment_vector()
            self.expanded = position
        

    def do_compress_hadamard_intern(self, position):
        """Auxiliary method method for method compress_hadamard()

        It executes method compress_hadamard() in case 
        position < self.LOG_INT_FIELDS.
        """
        assert 1 <= position < self.LOG_INT_FIELDS
        TOTAL_BITS = self.FIELD_BITS << position
        SH = TOTAL_BITS - self.FIELD_BITS
        LO_MASK = self.B0_MASK[1] & self.B0_MASK[1 << position] 
        HI_MASK = LO_MASK << self.FIELD_BITS
        for var in self.vars:
            self.assign(var, (var & LO_MASK) | ((var >> SH) & HI_MASK))
 

    def compress_hadamard_extern(self, v1, v2):
        """Auxiliary method method for method compress_hadamard()

        It operates on two variables v1, v2 of type uint_mmv_t. It
        revereses the effect of method expand_hadamard_extern().
        """
        self.assign_xor(v1, v2 << self.FIELD_BITS )

    def do_compress_hadamard_extern(self, position):
        """Auxiliary method method for method compress_hadamard()

        It executes method compress_hadamard() in case 
        position >= self.LOG_INT_FIELDS.
        """
        assert self.LOG_INT_FIELDS <= position <= self.LOG_VLEN
        s = ""
        j = 1 << (position - self.LOG_INT_FIELDS)
        mask = self.H_MASK_EXTERN
        for i in range(self.vlen):
            if i & j == 0:
                v1, v2 = self.vars[i], self.vars[i + j]
                self.compress_hadamard_extern(v1, v2)
        if position == self.LOG_VLEN:
            self.vlen >>= 1
            self.vars.resize(self.vlen)
        self.expanded = 0 


    def compress_hadamard(self):
        """Reverse the effect of method expand_hadamard().""" 
        if self.expanded: 
            if self.verbose:   
                print("compress vector after Hadamard operation")         
            self.comment("Reverse expansion for Hadamard operation")
            if self.expanded < self.LOG_INT_FIELDS:             
                self.do_compress_hadamard_intern(self.expanded)
            else:
                self.do_compress_hadamard_extern(self.expanded)
            self.expanded = 0
            self.comment_vector()
 

    def mul_var_pwr2(self, var, l):
        """Multiply all components of variable 'var' by scalar 2**l.

        Here l may be any integer.
        Modulo p = 2**k-1, this is simply a bit rotation.
        """
        ls = l % self.P_BITS
        if ls != 0 and self.P == 3:
            self.assign(var, ~var)
        elif ls != 0:
            hs = self.P_BITS - ls
            m_l = self.smask(range(hs))
            m_h = self.smask(range(hs, self.P_BITS))
            self.assign(var, ((var & m_l) << ls) | ((var & m_h) >> hs))

      

    def mul_pwr2(self, l):
        """Multiply vector v by the scalar 2**l.

        Here l may be any integer.
        Modulo p = 2**k-1, this is simply a bit rotation.
        """
        if l % self.P_BITS == 0:
            s = "Multiplication by 2**%d is trivial mod %d"
            self.comment(s % (l, self.P))
            return
        s = "Multiply vector by scalar 2**%d mod %d"
        self.comment(s % (l, self.P))
        for w in self.vars:
            self.mul_var_pwr2(w, l)
            


    def load_vector_direct(self, array_name):
        """Load vector v form an (external) array A of integers.

        'array_name' must a variable of type uint_mmv_t 
        referring to array A.
        """
        self.comment("Loading vector v from array %s" % array_name)
        for i,v in enumerate(self.vars):
            self.add("%s = %s[%d];\n" % (v, array_name, i), 1) 
        self.comment_vector()  

    def store_vector_direct(self, array_name):
        """Store a vector v to an (external) array A of integers.

        'array_name' must a variable of type uint_mmv_t 
        referring to array A.
        """ 
        self.comment("Storing vector v to array %s" % array_name)
        for i,v in enumerate(self.vars):
            self.add("%s[%d] = %s;\n" % (array_name, i, v), 1) 


    def complement_variable(self, var):
        """Generate code: 'var = ~var'; for variable 'var'"""
        self.add("%s = ~(%s);\n" % (var, var), 1, 1)

    def code(self, string):
        """Deprecated"""
        err = "Method 'code' has been removed, use 'add' instead!"
        raise AttributeError(err)
        #self.add(string)



    def generate(self):
        if self.verbose:
            print("Generate C code using class HadamardMatrixCode")
        """Return C code generated so far as a string"""
        s00 = """
// This is an automatically generated matrix operation, do not change!
{
"""     
        s01 = self.vars.declare() + "\n"
        s1 ="""}
// End of automatically generated matrix operation.
 
"""   
        s = "".join([s00, s01] + self.matrix_code + [s1]) 
        self.reset_vars()
        return s




from __future__ import absolute_import, division, print_function
#from __future__ import  unicode_literals


import sys
import os
from numbers import Integral

from mmgroup.bitfunctions import bitparity, bitweight
from mmgroup.dev.hadamard.hadamard_codegen import HadamardMatrixCode
from mmgroup.dev.hadamard.hadamard_codegen import C_UintVarPool
from mmgroup.generate_c import UserDirective, UserFormat



class C_UintVarArray(C_UintVarPool):
    """Model a (usually short) local array of integer variables

    The entries of the array can be treated in the same way as the
    variable of a pool, where the pool is an instance of class
    C_UintVarPool.
    """
    def __init__(self, c_type = "uint_mmv_t", name = "a[%d]", length = 0):
        """Create an array of integer variables of type 'ctype'.

        The array will contain 'length'  variables. These
        variables are instances of class C_UintVar and they are
        obtained as self[0], ..., self[length - 1].

        The names of the variables are name % 0, name % 1, name % 2, ..
        where name is given by parameter 'name'. So for obtaining
        a[0], a[1], a[2], ... parameter 'name' should be "a[%d]"
        """
        super(C_UintVarArray, self).__init__(c_type, name, length)
        self.array_name = self.name[:self.name.index("[")]
             

    def declare(self):
        """Declare the array in C

        The function returns a string containing that declaration.
        """
        if len(self) == 0:
            return ""
        name = self.array_name
        return "%s %s[%d];\n" % (self.c_type, name, len(self))

    def load_pool(self, start,  pool):
        """Load a 'pool' from the array of variables
 
        Here 'pool' is an instance of class C_UintVarPool
        modelling a pool of integer vriables. The function loads
        all 'official' variables of the pool from the array, 
        starting at the index 'start'.
        """
        if isinstance(start, Integral):
            indices = map(str, range(start, start + len(pool)))
        else:
            indices = ["(%s) + %d" % (start, i) for i in range(len(pool))]
        name = self.array_name
        for i, index in enumerate(indices):
            pool[i].assign("%s[%s]" %  (name, index)) 

    def store_pool(self, pool, start):
        """Store a 'pool' to the array of variables
 
        Here 'pool' is an instance of class C_UintVarPool
        modelling a pool of integer vriables. The function stores
        all 'official' variables of the pool to the array, starting 
        at the index 'start'.
        """
        if isinstance(start, Integral):
            indices = map(str, range(start, start + len(pool)))
        else:
            indices = ["(%s) + %d" % (start, i) for i in range(len(pool))]
        name = self.array_name
        for i, index in enumerate(indices):
            s = "%s[%s] = %s;\n" %  (name, index, pool[i])
            self.context.add(s, 1)

    def _bad(self, *args, **kwds):
        raise NotImplementedError

    temp = _bad
    xch = _bad


class HadamardOpXi64(HadamardMatrixCode):
    """Code for multiplying vector by a modified Hadamard matrix

    This class supports the operation of the  element l**e,
    e = 1,2  on a block of a vector v of length 64 with tag 'Y' 
    or 'Z' in  representation  196884x  of the monster group as 
    described in [Seys19].

    The group element l is called xi in [Seys19], but we cannot 
    use greek letters as tags in python or C.  We have xi**3 == 1. 
    Here a block of xi is a 64 times 64 matrix acting on a block 
    of v by right multiplication. For xi acting on a components
    of v with tags 'Y' and 'Z', we can partition xi  into blocks 
    of identical 64  times 64  matrices xi64, where each block 
    xi64  has shape xi64 = DD4 * M * DD16. Similarly, we can 
    partition  xi**2  into identical blocks xi64**2, with 
    xi64**2 = DD16 * M * DD4. Here  DD16 and DD4 are certain 
    diagonal matrices, and M is a Hadamard-like matrix as 
    described below. 

    We have DD16 = kron(D16, U4), DD4 = kron(U16, D4) with 
    D16 the diagonal matrix 
    ( -1, -1, -1, 1, -1, 1, 1, 1, -1, 1, 1, 1, 1, 1, 1, -1)
    and D4 the diagonal matrix (-1, 1, 1, 1), and U<n> the
    n times n unit matrix. kron() is the kronecker product
    as given by the numpy function numpy.kron().

    M is equal to kron(M4, M2), where M<n> is the parity-adjusted 
    2**n times 2**n Hadamard matrix  for an even number n as 
    defined in the header of module define_matrices.py. This class 
    supports the generation of C code for multiplying a vector v
    of integers modulo p by a matrix xi64 or xi64**2.

    Notation and implementation of vectors is as in the base 
    class HadamardMatrixCode, which supports right-multiplication 
    of a vector by a Hadamard  matrix.

    More precisely, M acts on each components of a vector labeled
    (U, i0 + i, j0 + j), i = 0,...,15, j = 0,..3, with column
    4 * i + j of matrix M acting on (U, i0 + i, j0 + j). Here the
    tag U must be 'Y' or 'Z', i0 must be divisible by 16, and j0
    must be divisible by 4.

    See  [Seys19], section 9 and module define_matrix.py for
    background.

    The code generated by method self.make_code() multiplies 16
    consecutive rows (U, i0 + i, j), i = 0,...,15, j = 0,..23,
    with 6 identical matrix blocks, where each of these blocks
    is either xi64 or x64**2. The code block generated by that
    method is rather large. On the other hand, this is one of the
    most time-critical operations in the monster group. Here 
    dividig a row (U, i, j), j = 0,...,23 into blocks of four
    entries is also quite time critical, requiring a fair amount 
    of optimization.
    """
    _PRE_MDIAG16 = [
            -1, -1, -1, 1, -1, 1, 1, 1, -1, 1, 1, 1, 1, 1, 1, -1
    ]
    MDIAG16 = [ int(i < 0) for i in _PRE_MDIAG16 ]

    def __init__(self, p, verbose = 0):
        """Create an instance of the class.

        Parameters are as in the base class HadamardMatrixCode.
        p is the modulus, which must be 2**k - 1, 2 <= k <= 8.
        """
        super(HadamardOpXi64, self).__init__(p, 0, verbose)
        # Next set sizes for Hadamard-like matrices
        self.set_sizes()
        # Next set the default position for method expand_hadamard    
        self.free_cy_pos = -1  
        # Next set positions where Hadamard operation muse  be done
        self.hadamard_operations = 3 + (15 << self.LOG_INT_FIELDS)
        # make tables and directives for code generation
        self.tables.update(self.make_tables())
        self.directives.update(self.make_directives())


    def set_sizes(self):
        array_size, log_vars = 0, 4
        if self.V24_INTS_USED == 1:
            array_size, log_vars = 24, 3
        elif self.V24_INTS_USED == 2:
            cy = self.NO_CARRY
            array_size, log_vars = 24 << cy, 3 + cy
        self.reset_vars(self.LOG_INT_FIELDS + log_vars)
        self.array = C_UintVarArray(length = array_size) 
        self.array.set_context(self)   

    def load_var(self, source, i, ofs, p_mask, dest, tmp = None):
        if tmp is None:
            tmp = self.vars.temp()
        i1 = 15 - i if bitparity(i) else i
        index = (i1 << self.LOG_V24_INTS) + ofs
        mindex = 2 * self.MDIAG16[i1] 
        dest.assign("%s[%d] ^  %s[%s]" % (source, index, p_mask, mindex))
        sh = self.FIELD_BITS
        swap_mask = self.smask(self.P, range(1, 24, 4))
        tmp.assign((dest ^ (dest >> sh)) & swap_mask)
        dest.assign_xor(tmp | (tmp << sh))


    def store_var(self, source, p_mask, dest, i, ofs, tmp = None):
        if tmp is None:
            tmp = self.vars.temp()
        index = (i << self.LOG_V24_INTS) + ofs
        mindex = 2 * self.MDIAG16[i] + 4
        s = ""
        if self.V24_INTS_USED < 3 and ofs == self.V24_INTS_USED - 1:
            mask1 = self.hex(self.smask(self.P, range(24 % self.INT_FIELDS)))
            s += "%s[%d] = (%s ^  %s[%s]) & %s;\n" % (
                dest, index, source, p_mask, mindex, mask1)
            self.n_operations += 1
        else:
            s += "%s[%d] = %s ^  %s[%s];\n" % (
                dest, index, source, p_mask, mindex)
        self.add(s, 1, 1)

    def load_all_v24_1(self, source, p_mask):
        t = self.vars.temp(0)
        for i in range(16):
            v0 = self.vars[i]
            self.load_var(source, i, 0, p_mask, v0, t)
        self.hadamard_operations = 3 + (15 << 5)

    def store_all_v24_1(self, p_mask, dest):
        t = self.vars.temp(0)
        for i in range(16):
            v0 = self.vars[i]
            #self.mul_var_pwr2(v0, -3) 
            self.store_var(v0, p_mask, dest, i, 0, t)   


    def load_all_v24_1_no_cy(self, source, p_mask):
        t, t1 = self.vars.temp(0), self.vars.temp(1)
        sh1 = 8 * self.FIELD_BITS
        sh2 = 16 * self.FIELD_BITS
        m0, m1, m2 = [self.smask(self.P, range(8*i, 8*i+8)) for i in range(3)]
        for i in range(8):
            v0, v1, v2 = self.vars[i], self.array[i+8], self.array[i+16]
            self.load_var(source, 2*i, 0, p_mask, v0, t1)
            self.load_var(source, 2*i+1, 0, p_mask, t, t1)
            v1.assign(((v0 >> sh1) & m0) | ((t << sh1) & m2))
            v2.assign(((v0 >> sh2) & m0) | ((t) & m2))
            v0.assign(((v0) & m0) | ((t << sh2) & m2))
        self.free_cy_pos = 3
        self.hadamard_operations = 3 + (15 << 4)
            
    def store_all_v24_1_no_cy(self, p_mask, dest):
        t, t1 = self.vars.temp(0), self.vars.temp(1)
        sh1 = 8 * self.FIELD_BITS
        sh2 = 16 * self.FIELD_BITS
        m0, m1, m2 = [self.smask(self.P, range(8*i, 8*i+8)) for i in range(3)]
        for i in range(8):
            v0, v1, v2 = self.array[i], self.array[i+8], self.vars[i]
            t.assign((v0 & m0) | ((v1 << sh1) & m1) | (v2 << sh2))
            #self.mul_var_pwr2(t, -3)           
            self.store_var(t, p_mask, dest, i+i, 0, t1)   
            v0.assign(((v0 >> sh2) & m0) | ((v1 >> sh1) & m1) | (v2 & m2))
            #self.mul_var_pwr2(v0, -3)           
            self.store_var(v0, p_mask, dest, i+i+1, 0, t1)


    def load_all_v24_2(self, source, p_mask):
        t, t1 = self.vars.temp(0), self.vars.temp(1)
        sh1 = 8 * self.FIELD_BITS
        m0, m1 = [self.smask(self.P, range(8*i, 8*i+8)) for i in range(2)]
        for i in range(8):
            v0, v1, v2 = self.vars[i], self.array[i+8], self.array[i+16]
            self.load_var(source, 2*i, 0, p_mask, v0, t1)
            self.load_var(source, 2*i+1, 0, p_mask, t, t1)
            v1.assign(((v0 >> sh1) & m0) | ((t) & m1))
            v0.assign(((v0) & m0) | ((t << sh1) & m1))
            self.load_var(source, 2*i, 1, p_mask, v2, t1)
            self.load_var(source, 2*i+1, 1, p_mask, t, t1)
            v2.assign(((v2) & m0) | ((t << sh1) & m1))
        self.hadamard_operations = 3 + (15 << 3)
            

    def store_all_v24_2(self, p_mask, dest):
        t, t1 = self.vars.temp(0), self.vars.temp(1)
        sh1 = 8 * self.FIELD_BITS
        m0, m1 = [self.smask(self.P, range(8*i, 8*i+8)) for i in range(2)]
        for i in range(8):
            v0, v1, v2 = self.array[i], self.array[i+8], self.vars[i]
            t.assign((v0 & m0) | ((v1 << sh1) & m1)) 
            #self.mul_var_pwr2(t, -3)           
            self.store_var(t, p_mask, dest, 2*i, 0, t1)   
            v0.assign(((v0 >> sh1) & m0) | ((v1) & m1))
            #self.mul_var_pwr2(v0, -3)           
            self.store_var(v0, p_mask, dest, 2*i+1, 0, t1)
            #self.mul_var_pwr2(v2, -3)           
            t.assign((v2) & m0)
            self.store_var(t, p_mask, dest, 2*i, 1, t1)
            v2.assign((v2 >> sh1) & m0)
            self.store_var(v2, p_mask, dest, 2*i+1, 1, t1)


    def load_all_v24_2_no_cy(self, source, p_mask):
        t = self.vars.temp(0)
        sh1 = 8 * self.FIELD_BITS
        m0, m1 = [self.smask(self.P, range(8*i, 8*i+8)) for i in range(2)]
        for i in range(16):
            v0, v1, v2 = self.vars[i], self.array[i+16], self.array[i+32]
            self.load_var(source, i, 0, p_mask, v0, t)
            v1.assign((v0 >> sh1) & m0)
            v0.assign((v0) & m0)
            self.load_var(source, i, 1, p_mask, v2, t)
        self.hadamard_operations = 3 + (15 << 4)
        self.free_cy_pos = 3

    def store_all_v24_2_no_cy(self, p_mask, dest):
        t = self.vars.temp(0)
        sh1 = 8 * self.FIELD_BITS
        m0, m1 = [self.smask(self.P, range(8*i, 8*i+8)) for i in range(2)]
        for i in range(16):
            v0, v1, v2 = self.array[i], self.array[i+16], self.vars[i]
            v0.assign((v0 & m0) | ((v1 << sh1) & m1))
            v2.assign(v2)
            #self.mul_var_pwr2(v0, -3)           
            #self.mul_var_pwr2(v2, -3)           
            self.store_var(v0, p_mask, dest, i, 0, t)
            self.store_var(v2, p_mask, dest, i, 1, t)

    def small_load_all(self, source, p_mask):
        if self.V24_INTS_USED == 1:
            if self.NO_CARRY:
                self.load_all_v24_1_no_cy(source, p_mask)
            else:
                raise NotImplmentedError("V24_INTS_USED=1")
        elif self.V24_INTS_USED == 2:
            if self.NO_CARRY:
                self.load_all_v24_2_no_cy(source, p_mask)
            else:
                self.load_all_v24_2(source, p_mask)
        else:
            raise NotImplmentedError("V24_INTS_USED > 2")
  

    def small_store_all(self, p_mask, dest):
        if self.V24_INTS_USED == 1:
            if self.NO_CARRY:
                self.store_all_v24_1_no_cy(p_mask, dest)
            else:
                raise NotImplmentedError("V24_INTS_USED=1")
        elif self.V24_INTS_USED == 2:
            if self.NO_CARRY:
                self.store_all_v24_2_no_cy(p_mask, dest)
            else:
                self.store_all_v24_2(p_mask, dest)
        else:
            raise NotImplmentedError("V24_INTS_USED > 2")
         

    def add_mask(self, p_mask, value):
        d = {1: "++", -1: "--"}
        self +=  "%s%s;\n" %(d[value], p_mask)
        
    def label(self, index):
        return "l_mmv%d_op_l64_%d"  % (self.P, index)


    def very_small_op(self, source, p_mask, dest):
        self.reset_vars(4+5)
        self.load_all_v24_1(source, p_mask)
        self.comment_statistics()
        self.hadamard_op(self.hadamard_operations, shift = -3)
        self.comment_statistics()
        self.store_all_v24_1(p_mask, dest)
        self += "%s += %d;\n" % (source, 16 * self.V24_INTS)
        if dest != source:
            self += "%s += %d;\n" % (dest, 16 * self.V24_INTS)
        self.comment_statistics()


        
    def small_op(self, source, p_mask, dest):
        self.small_load_all(source, p_mask)
        l1, l2 = self.label(1), self.label(2)
        self.comment_statistics()
        a = self.array
        self += "i = 0;\ngoto %s;\n" % l2
        self += l1 + ":\n"
        a.store_pool(self.vars, "i")
        self += "i += %d;\n" % len(self.vars)
        a.load_pool("i",  self.vars)
        self += l2 + ":\n"
        self.expand_hadamard(self.free_cy_pos)   
        self.hadamard_op(self.hadamard_operations, shift = -3, 
            compress = True)
        #self.compress_hadamard()   
        self.comment_statistics()
        i_end = 2 * len(a) // 3
        self += "if (i < %d) goto %s;\n" % (i_end, l1)
        self.small_store_all(p_mask, dest)
        self += "%s += %d;\n" % (source, 16 * self.V24_INTS)
        if dest != source:
            self += "%s += %d;\n" % (dest, 16 * self.V24_INTS)
        self.comment_statistics()

    def large_op_on_int(self, source, p_mask, dest):
        t = self.vars.temp(0)
        for i in range(16):
            self.load_var(source, i, 0, p_mask, self.vars[i], t)    
        self.hadamard_op(self.hadamard_operations, shift = -3)
        for i in range(16):
            #self.mul_var_pwr2(self.vars[i], -3)           
            self.store_var(self.vars[i], p_mask, dest, i, 0, t)
        self += "%s++;\n" % source
        if dest != source:
            self += "%s++;\n" % dest


    def large_op(self, source, p_mask, dest):
        USED = self.V24_INTS_USED
        self += "for (i = 0; i < %d; ++i) {\n" % USED
        self.large_op_on_int(source, p_mask, dest)
        self += "}\n"
        for i in range(16):
            for j in range(self.V24_INTS - USED):
                k = i * self.V24_INTS + j
                self += "%s[%d] = 0;\n" % (dest, k)
        k = 16 * self.V24_INTS - USED
        self += "%s += %d;\n" % (source, k)
        if dest != source:
            self += "%s += %d;\n" % (dest, k)


    def additional_declare(self):
        self += self.array.declare()
        self += ("uint_fast32_t i;\n\n")

    def make_code(self, source, p_mask, dest):
        """Apply operation on vector.

        YET TO BE DOCUMENTED!!!

        Right multiply the vector of integers mod self.P stored in 
        'source' by t**e, where t is the 64 times 64 triality matrix 
        operating on blocks of the rep 196884x with tag. 


        Let A_T be the table created by function  self.exp_table(). 
        'p_mask' must be a pointer of type uint_mmv_t* with value
        p_mask = &A_T[e-1] for e = 1, 2.
         
        'source' and 'dest' must be pointers of type uint_mmv_t*. 
        """
        self.reset_vars()
        if not self.FAST_MOD3:
            self.additional_declare()
        self.comment(
"""TODO: write comment!!!
""".format(p = self.P, source = source, p_mask = p_mask, dest = dest)
        )
        if self.V24_INTS_USED < 3:
            if self.FAST_MOD3:
                self.very_small_op(source, p_mask, dest)
            else:
                self.small_op(source, p_mask, dest)
        else: 
            self.large_op(source, p_mask, dest)  
        return self.generate()

    def exp_table(self):
        d4 = self.smask(self.P, range(0, 24, 4))
        d16 = self.smask(self.P, (1 << 24) - 1)
        m = [0] * 8
        m[0:8:2] = [d4, d4, 0, d16] 
        m[1:8:2] = [0, d16, d4, d4] 
        return m


    def make_tables(self):
        return {
            "TABLE_MUL_MATRIX_XI64" : self.exp_table()
        }


    def make_directives(self):
        return {
            "MUL_MATRIX_XI64" : UserDirective(self.make_code, "sss"),
        }
        


class HadamardOpXi16(HadamardMatrixCode):

    def __init__(self, p, verbose = 0):
        """Create an instance of the class.

        Parameters are as in the base class HadamardMatrixCode.
        p is the modulus, which must be 2**k - 1, 2 <= k <= 8.
        """
        super(HadamardOpXi16, self).__init__(p, 0, verbose)
        # Next set sizes for Hadamard-like matrices
        self.reset_vars(self.LOG_INT_FIELDS + 2)
        # Next set positions where Hadamard operation muse  be done
        self.hadamard_operations = 3 + (3 << self.LOG_INT_FIELDS)
        # set masks for diagonal operation
        self.MASKS = [self.smask(self.P, range(0,24,4))]
        self.MASKS.append(~self.MASKS[0] & self.smask(self.P, range(24)))
        # make tables and directives for code generation
        #self.tables.update(self.make_tables())
        self.directives.update(self.make_directives())
        

    def load_var(self, source, i, mask, dest):
        tmp = self.vars.temp()
        i1 = [0, 2, 1, 3][i]
        index = (i1 << self.LOG_V24_INTS)
        m = self.hex(self.MASKS[i1 == 0])
        dest.assign("%s[%d] ^  (%s & %s)" % (source, index, m, mask))
        sh = self.FIELD_BITS
        swap_mask = self.smask(self.P, range(1, 24, 4))
        tmp.assign((dest ^ (dest >> sh)) & swap_mask)
        dest.assign_xor(tmp | (tmp << sh))


    def store_var(self, source, mask, dest, i):
        index = (i << self.LOG_V24_INTS)
        m = self.hex(self.MASKS[i == 0])
        s =  "%s[%d] = %s ^ (%s & %s);\n" % (
            dest, index, source, mask, m)
        self.add(s, 1, 2)



    def op(self, source, mask, dest):
        self.complement_variable(mask)
        for i in range(4):
            self.load_var(source, i, mask, self.vars[i])
        self.hadamard_op(self.hadamard_operations)
        self.complement_variable(mask)
        for i in range(4):
            self.mul_var_pwr2(self.vars[i], -2)           
            self.store_var(self.vars[i], mask, dest, i)  
        self.comment_statistics()
        self += "%s++;\n" % source
        if dest != source:
            self += "%s++;\n" % dest


    def make_code(self, source, mask, dest):
        """Apply operation on vector.

        YET TO BE DOCUMENTED!!!

        Right multiply the vector of integers mod self.P stored in 
        'source' by t**e, where t is the 64 times 64 triality matrix 
        operating on blocks of the rep 196884x with tag. 

        For e = 1, parameter 'mask' must be 0; for e = 2, parameter
        'mask' must bei (uint_mmv_t)(-1). The result is stored
        in 'dest'.

        'source' and 'dest' must be pointers of type uint_mmv_t*,
        'mask' must be a variable of type uint_mmv_t.
        """
        self.reset_vars()
        USED = self.V24_INTS_USED
        if USED > 1:
            self += "uint_fast32_t i;\n"
        self.comment(
"""TODO: write comment!!!
""".format(p = self.P, source = source, mask = mask, dest = dest)
        )
        if USED > 1:
            self += "for (i = 0; i < %d; ++i) {\n" % USED
        self.op(source, mask, dest)
        if USED > 1:
            self += "}\n" 
        if USED == 2:
            for i in range(-1, 7, 2):
                mask1 = self.hex(self.smask(self.P, range(8)))
                self.add("%s[%s] &= %s;\n" % (dest, i, mask1), 1, 1)
        elif USED >= 3:
            for i in range(4):
                for j in range(self.V24_INTS - USED):
                    k = i * self.V24_INTS + j
                    self += "%s[%d] = 0;\n" % (dest, k)
        k = 4 * self.V24_INTS - USED
        self += "%s += %d;\n" % (source, k)
        if dest != source:
            self += "%s += %d;\n" % (dest, k)
        return self.generate()

    def make_directives(self):
        return {
            "MUL_MATRIX_XI16" : UserDirective(self.make_code, "sss"),
        }
  


if __name__ == "__main__":
    for p in (3,7, 15, 127):
        print( """typedef unsigned int uint_fast32_t;
typedef unsigned long long  uint_mmv_t;
void mm_f%d(uint_mmv_t *source, uint_mmv_t mask, uint_mmv_t *dest)
""" % p )
        cg = HadamardOpXi16(p)
        print("\n", cg.make_code("source", "mask", "dest"))



from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

import sys
import os
import re
import warnings

from mmgroup import mat24
from mmgroup.dev.mm_op.mm_op import MM_Op


###########################################################################
# Table for operation xy on the 64 entries of for tag T
###########################################################################

class Perm64_xy(MM_Op):
    """Yet to be documented!!!!!!!!!!!!!!!!!!!!!

    """
    weights = sum(mat24.suboctad_weight(j) << j for j in range(64))
    mask = (1 << 64) - 1
    hi_list = [ 0, mask, weights, weights ^ mask ]
    directives = {}

    def __init__(self, p):
        """Initialise for calulations with small integers modulo p

        p+1 must be a power of two. Calculations modulo p are described 
        in more detail in the base classes of this class.
        """
        super(Perm64_xy, self).__init__(p)
        self.t_hi = self.hi_table()
        self.t_lo = self.lo_table()
        self.tables.update(self.make_tables())

    def table_value(self, index):
        i0 = index & 0x3f
        v = sum(mat24.suboctad_scalar_prod(j, i0) << j for j in range(64))
        return self.hi_list[ (index >> 6) & 3 ] ^ v

    def table_part(self, index):
        tbl = self.table_value(index)
        p, i_fields = self.P, self.INT_FIELDS
        return [self.smask(p, tbl >> i) for i in range(0, 64, i_fields)]

    def hi_table(self):
        return sum((self.table_part(i) for i in range(0, 256, 16)), [])
        
    def lo_table(self):
        return sum((self.table_part(i) for i in range(16)), [])

    def make_tables(self):
        return {
            "TABLE_PERM64_XY_LOW" : self.t_lo,
            "TABLE_PERM64_XY_HIGH" : self.t_hi,
        }
 





from numbers import Integral, Number
from random import randint


#######################################################################
# Class Parity 
#######################################################################


class Parity():
    """This class models the parity of a bit vector.

    As one might expect, the parity of a bit vector may be odd or 
    even, and it is an element of the field ``GF(2)`` of two elements.

    This means that arithmetic operations ``+``, ``-``, and ``*`` 
    work as expected on instances of class |Parity|, or on 
    instances of this class combined with integers. 

    The values ``1**p`` and ``(-1)**p`` are defined as usual for an
    object ``p`` of class |Parity|.
    Similarly, ``g**p`` is defined  if ``g`` is an element of a group 
    defined in this package and has order ``1`` or ``2``. More precisely, 
    ``g`` must be an instance of  class ``AbstractGroupWord`` here.

    Bitwise operations ``&``, ``|``, and ``^`` are illegal on  
    instances of class  |Parity|.

    ``Parity(x)`` is defined if ``x`` an integer, an instance of class
    |Parity|, or an instance of any class defined is this package which 
    has a natural  parity, e.g. an instance of class |GcVector| 
    or |Cocode|.

    ``int(p)`` evaluates to ``0`` or ``1`` for an instance of 
    class |Parity|.


    Implementation remarks
    
      If an object ``x`` has an attribute or a property ``parity`` then
      ``Parity(x)`` means ``Parity(x.parity)`` and ``x + p`` means 
      ``Parity(x) + p`` for any  instance ``p`` of class Parity.

      If an object ``x`` has an attribute or a property ``parity_class`` 
      then  ``x * p``  and  ``p * x``  mean  
      ``Parity(parity_class(x) * int(p))``. If that attribute or a 
      property has value ``True`` then ``x * p``  and  ``p * x``  mean  
      ``Parity(x * int(P))``.
    """
    __slots__ = "value"
    names = ["<even>", "<odd>"]
    ERR_MOD2 = "%s object may be reduce mod 2 only"
    ERR_PARITY = "%s object does not have a parity"

    def __init__(self, value):
        if isinstance(value, Integral):
            self.value = value & 1
        elif isinstance(value, Parity):
            self.value = value.value & 1
        elif isinstance(value, str):
            self.value = randint(0, 1)
        else:
            try:
                self.value = int(value.parity) & 1
            except:
                raise TypeError(self.ERR_PARITY % type(value)) 

    def __int__(self):
        return self.value & 1

    @property
    def ord(self):
        """Return ``1`` if the parity is odd and ``0`` if it is even"""
        return self.value

    def __eq__(self, other):
        return (isinstance(other, Parity)  
            and (self.value ^ other.value) & 1 == 0)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __pos__(self):
        return self
        
    __neg__ = __pos__


    def __add__(self, other):
        if isinstance(other, Integral):
            return Parity((self.value + other) & 1)
        elif isinstance(other, Parity):
            return Parity((self.value + other.value) & 1)
        else:
            try:
                parity = other.parity
            except:
                return NotImplemented
            else:
                return Parity((self.value + parity) & 1)

    __sub__ = __add__
    __radd__ = __add__
    __rsub__ = __radd__

    def __mul__(self, other):
        if isinstance(other, Integral):
            return Parity(self.value & other & 1)
        elif isinstance(other, Parity):
            return Parity(self.value & other.value & 1)
        else:
            try:
                parity_class = other.parity_class
            except AttributeError:
                return NotImplemented
            else:
                if parity_class is True:
                   return other * self.value
                else:
                   return parity_class(other) * self.value

    __rmul__ = __mul__


    def __truediv__(self, other):
        if isinstance(other, Integral):
            if abs(other) == 1:
                return self
        else:
            return NotImplemented

    def __mod__(self, other):
        if isinstance(other, Integral):
            if other == 2:
                return self
            else:
                raise ValueError(ERR_MOD2 % type(self))
        else:
            return NotImplemented


    def __rpow__(self, other):
        if isinstance(other, Integral):
            if other == 1:
                return 1
            elif other == -1:
                return (-1) ** self.value
            else:
                err = "Basis must be +-1 for type Parity exponent"
                raise ValueError(err)
        else:
            return NotImplemented
 
    def __str__(self):
        return self.names[self.value & 1]

    __repr__ = __str__






import numpy as np

from mmgroup.mat24 import m24num_to_perm
from mmgroup.mat24 import gcode_to_vect


def mul24_perm(v, pi):
    perm = m24num_to_perm(pi) 
    w = np.zeros(24, dtype = np.int32)
    for i in range(24):
        w[perm[i]] = v[i]
    for i in range(24):
        v[i] = w[i]

def mul24_xi_sym(v):
    parity = 0 
    for i in range(0, 24, 4):
        s = v[i] + v[i+1] + v[i+2] + v[i+3]
        parity |= s
        s >>= 1
        v[i] -= s
        v[i+1] -= s
        v[i+2] -= s
        v[i+3] -= s
    if parity & 1:
        raise ValueError("Operation on Leech space is not integral")

def mul24_xi_diag(v):
    for i in range(0, 24, 4):
        v[i] = -v[i]

def mul24_xi(v, e):
    e = e % 3
    if e == 1:
        mul24_xi_sym(v)
        mul24_xi_diag(v)
    elif e == 2:
        mul24_xi_diag(v)
        mul24_xi_sym(v)
  

def mul24_y(v, y):
    y1 = gcode_to_vect(y)
    for i in range(24):
        if (y1 >> i) & 1:
            v[i] = -v[i]

def mul24_id(v, _):
    pass

def mul24_t(v, _):
    err = "Operation of triality element not defined on Leech lattice"
    raise ValueError(err)


g_functions = {
   'd':  mul24_id,
   'p':  mul24_perm,
   'x':  mul24_id,
   'y':  mul24_y,
   'l':  mul24_xi,
   't':  mul24_t,
}


def mul24_mmgroup(v, g):
    """Multiply vector in Leech space with an element of the monster

    Here ``v`` should be a numpy array of ``dtype = np.int32`` and 
    ``shape = (24,)``. ``v`` represents an vector in the underlying
    space of the Leech lattice.
   
    ``g`` is an instance of class |MMGroup| representing an element of
    the monster group. Actually, ``g`` should be in the subgroup of
    the monster generated by tags, ``d, p, x, y, z, l``.

    The function computes ``v = v * g`` inplace and returns the 
    modified vector ``v``.
    
    The function may half some entries of vector v. If the result
    in not a vector of integers, the function raises ValueError.

    Caution:
    
    The operation of ``g`` on ``v`` is defined up to sign only!
    """
    v = np.copy(v)
    for tag, a in g.as_tuples():
        g_functions[tag](v, a)
    return v




"""The module computes a table for the Parker Loop cocycle ``theta``.

"""


from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


from random import randint
import types
import sys
import re
import os
from operator import __or__, __xor__

import numpy
from numpy import array, zeros, uint8, uint16, uint32, uint64


from mmgroup.bitfunctions import bw24, v2

from mmgroup.dev.mat24.mat24tables import Mat24Tables


from mmgroup.dev.mat24.mat24aux import split_golay_codevector





def theta_to_basis_vector(v):
    """Return suitable theta(v) value for Golay code basis vector v

    Here theta(v) is a cocode word, which is (loosely speeking)
    the cocycle of the Parker loop. Both, v and theta(v) are 
    vectors given in 'vect' representation.

    We assume that vector v is either 'grey' or 'colored' in the
    sense of [Seys19]. If v is a 'grey' vector, we return the
    cocycle theta(v) as given by Lemma 3.9. For each 'colored'
    vasis vector v, we assume that theta(v) is 'grey'; then the
    'grey' part of theta(v) is uniquely defined by Lemma 3.9.  
    """
    bw, col = split_golay_codevector(v, check=Mat24Tables)
    if col and not bw:
        assert bw24(col) == 8
        return ((col >> 1) | (col >> 2) | (col >> 3)) & 0x111111
    assert col == 0 
    bw &= 0x111111
    weight = bw24(bw)
    if weight < 2 or weight > 4:   return 0
    if weight == 2: return bw ^ 0x111111
    if weight == 4: return bw
    if weight == 3: return 0x111111  
    
    
def theta_basis_vectors(verbose = 0):
    """Return cocycle of the standard basis of the Golay code.

    Here the standard basis of the Golay code is given by
    Mat24Tables.basis[12:]. We return list containing theta(b_i) 
    for the basis vectors b_0,...,b_11. Here all values 
    theta(b_i) are computed by function theta_to_basis_vector(v).

    We also check that these values theta(b_i) are consisten.
    """
    basis = Mat24Tables.basis[12:]
    thetas = [theta_to_basis_vector(v) for v in basis] 
    for i in range(12):
        for j in range(i):
            cocycle = bw24(thetas[i] & basis[j]) 
            cocycle = (cocycle  ^ (bw24(thetas[j] & basis[i]))) & 1
            expected = (bw24(basis[i] & basis[j]) >> 1) & 1
            assert cocycle == expected & 1, (i, j, cocycle, expected)
        cocycle = bw24(thetas[i] & basis[i]) & 1
        assert cocycle == (bw24(basis[i]) >> 2) & 1
    if verbose:
        print( "Cocycle test for basis of Parker loop passed" )
    return thetas



def make_theta_table():
    r"""Creates the theta function for the Parker Loop
    theta() is a function from the Golay code C to the Golay cocode C*.

    Multiplication '(*)' in the Parker loop is defined by:

           a (*) b = (a ^ b) * (-1) ** scalar(theta(a),b) ,

    Where '^' is the vector addition in GF(2)**12, and  scalar(.,.)
    is the scalar product defined on C x C*.     

    More specifically, theta is a quadratic form from C onto C*. The 
    basis cc_i, i=0,...,11 of the cocode defined here is reciprocal to 
    the basis c_i, i=0,...,11 of the cocode defined here. 
    Define   theta(i)[j] = theta(i) * c_j.   Then 

            theta(i)  = sum[1 << theta(i)[j] for j in range(12) ]. 
 
    To define theta as a quadratic form, it suffices to define theta 
    on the basis vectors. We define theta on the basis vectors as
    given by function theta_to_basis_vector(v).

    There is a symmetric bilinear form B : from C x C onto C*
    associated with theta on all pairs of basis vectors satisfying:

             B(x,y) = theta(x+y) + theta(x) + theta(y)   .        
             B(x,y) = x :math:`\cap` Y

    Here :math:`\cap means intersection (i.e. bitwise 'and' in 
    GF(2)**24) of two Golay code words. This allows us to compute 
    theta(x), for all Golay code words, if theta is given on a 
    basis of the Golay code.
    """
    assert Mat24Tables.gcode_to_vect(0x800) == 0xffffff
    theta_table = numpy.zeros(0x800, dtype = uint16)

    thetas = theta_basis_vectors()
    for i in range(11):
         theta_table[1 << i] = Mat24Tables.vect_to_cocode(thetas[i])
 
    for i in range(0x800):
        if (i & (i-1)) != 0:    # i.e. if i is not 0 or a power of 2
            i0 = 1 << v2(i) 
            i1 = i ^ i0
            cap = Mat24Tables.gcode_to_vect(i0) 
            cap &= Mat24Tables.gcode_to_vect(i1)
            cc = Mat24Tables.vect_to_cocode(cap)
            theta_table[i] = theta_table[i0] ^ theta_table[i1] ^ cc
    return theta_table




def augment_theta_table(theta_table):
    """store bitweight of Golay code word i in table[i/2], bits 14..12"""
    for i in range(0x800):
        w = bw24(Mat24Tables.gcode_to_vect(i))
        assert w in [0,8,12,16]
        theta_table[i] |=  (w >> 2) << 12
    return  theta_table


def make_augmented_theta_table():
    theta_table = make_theta_table()
    return augment_theta_table(theta_table)
            



    

def make_autpl_qf_table(theta_table, bitwidth=32):
    qf_table = [theta_table[(1 << i)] & 0x7ff for i in range(11)]
    mask_table =  [(1 << i) - 1 for i in range(11)]
    if  bitwidth==32:
        tab = numpy.zeros(10, dtype = uint32)
        for i in range(0, 10, 2):
            tab[i] = qf_table[i+1] + (qf_table[i+2] << 16) 
            tab[i+1] = mask_table[i+1] + (mask_table[i+2] << 16) 
        return tab
    if bitwidth == 64:
        t = [sum((int(qf_table[5*i+j+1]) << (12*j) for j in range(5)))
            for i in range(2)]
        return ["0x%xULL" % x for x in t]
    raise ValueError("Illegal  bitwidth for table 'qf_table'")
        






# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
import subprocess
import sphinx_rtd_theme

sys.path.insert(0, os.path.abspath(os.path.join('..', '..')))
sys.path.insert(0, os.path.abspath(os.path.join('..', '..', 'src')))


# -- Check if we are in readthedocs -------------------------------------------

on_readthedocs = os.environ.get('READTHEDOCS') == 'True'


# -- Creating C files in readthedocs ----------------------------------------

SETUP_DIR = os.path.abspath(os.path.join('..', '..'))
C_DIR = os.path.join(SETUP_DIR, "src", "mmgroup", "dev", "c_files")
DOXYGEN_DIR = os.path.abspath(os.path.join('..', 'doxygen'))

if on_readthedocs:
    print("\nGenerating C files ...")
    subprocess.check_call([sys.executable, "setup.py", "build_ext"], 
        cwd=SETUP_DIR)
    print("C files have been generated\n")

print("\nDoxygen Directory = ", DOXYGEN_DIR)
print("C Directory =", C_DIR, ":")
C_FILES = [f for f in os.listdir(C_DIR) 
    if os.path.splitext(f)[1] in [".c", ".h"]]
print(C_FILES, "\n")


# -- Call doxygen ------------------------------------------------------

print(" \nStarting doxygen ...")
subprocess.check_call("doxygen", cwd = DOXYGEN_DIR)
print("End of doxygen\n")


# -- Project information -----------------------------------------------------

project = 'mmgroup'
copyright = '2020, Martin Seysen'
author = 'Martin Seysen'

# The full version, including alpha/beta/rc tags
release = '0.0.1'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
   'sphinx_rtd_theme',
   'sphinx.ext.autodoc',
   'sphinxcontrib.bibtex',
    'breathe',
]


bibtex_bibfiles = ['references.bib']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"



# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Sidebars in tables in html are ugly! Here we disable them: 
def setup(app):
    #app.add_stylesheet('custom.css') # appears to be deprecated!
    app.add_css_file('custom.css') 

# -- Options for latex output ------------------------------------------------
# The following option should remove excessive blank pages.

latex_elements = {
  'extraclassoptions': 'openany,oneside'
}



# -- Mock up Cython extension modules ----------------------------------------

# We will mock up most Cython extensions when generating the documentation
# with Sphinx. This means that these extensions need not be present when
# the documentation is generated. For background, see:
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#confval-autodoc_mock_imports

autodoc_mock_imports = [
  #  "mmgroup.mat24",   # We cannot mock up this extension, see below
    "mmgroup.generators",
    "mmgroup.clifford12",
    "mmgroup.mm",
    "mmgroup.mm3",
    "mmgroup.mm7",
    "mmgroup.mm15",
    "mmgroup.mm31",
    "mmgroup.mm63",
    "mmgroup.mm127",
    "mmgroup.mm255",
    "mmgroup.mm_reduce",
    "blah",
]

# For documentation, we cannot mock up the 'mmgroup.mat24' extension,
# since we need some data from that extension.
# Whenever needed, we will import a pure python substitute for the
# 'mmgroup.mat24' extension if the original extension has not been found.
# Class 'Mat24' in module  'mmgroup.dev.mat24.mat24_ref' is the
# best substitute available for 'mmgroup.mat24'.





# -- Workaround for a certain LaTeX error when building the .pdf file --------

# When using nested lists in Sphinx, the following error occurs when building
# the .pdf file:
# ! LaTeX Error: Too deeply nested.
# Here we implement a workaround for this problem, which has been taken from:
# https://stackoverflow.com/questions/28454217/how-to-avoid-the-too-deeply-nested-error-when-creating-pdfs-with-sphinx


latex_elements = {
# Additional stuff for the LaTeX preamble.
'preamble': r'\usepackage{enumitem}\setlistdepth{99}',
}




# -- Breathe Configuration --------

breathe_projects = {
   'mmgroup' : os.path.join(DOXYGEN_DIR, 'xml')
}

breathe_default_project = "mmgroup"


breathe_projects_source = {
    "mmgroup" :  (C_DIR, C_FILES) # C_FILES not yet known
}



